""" Functions for reading in MaxQuant search results.
"""
import pandas as pd
import polars as pl

from inspire.constants import (
    ACCESSION_KEY,
    CHARGE_KEY,
    DELTA_SCORE_KEY,
    ENGINE_SCORE_KEY,
    KNOWN_PTM_WEIGHTS,
    LABEL_KEY,
    MASS_DIFF_KEY,
    PEPTIDE_KEY,
    PTM_ID_KEY,
    PTM_IS_VAR_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
    PTM_WEIGHT_KEY,
    RT_KEY,
    SCAN_KEY,
    SEQ_LEN_KEY,
    SOURCE_KEY,
)
from inspire.utils import filter_for_prosit

MQ_ABBREVIATIONS = {
    'ox': 'Oxidation (M)',
    'ac': 'Acetyl (N-term)',
    'dm': 'Deamidated',
    'cm': 'Carbamidomethylation',
    'tm': 'TMT 6-plex',
}

ID_NUMBERS = {
    'unknown': 9,
    'Deamidation (N)': 5,
    'Deamidation (Q)': 5,
    'Phospho (S)': 4,
    'Phospho (T)': 4,
    'Phospho (Y)': 4,
    'Phospho(R)': 4,
    'Acetyl (N-term)': 3,
    'Acetyl (Protein N-term)': 3,
    'Oxidation (M)': 2,
    'Carbamidomethyl (C)': 1,
    'Carbamidomethylation': 1,
}

# Define the relevant column names from MaxQuant search results.
MQ_ACCESSION_KEY = 'Proteins'
MQ_CHARGE_KEY = 'Charge'
MQ_DECOY_KEY = 'Reverse'
MQ_DELTA_SCORE_KEY = 'Delta score'
MQ_LEN_KEY = 'Length'
MQ_MASS_KEY = 'Mass'
MQ_MISSED_CLEAVAGES_KEY = 'Missed cleavages'
MQ_MOD_SEQ_KEY = 'Modified sequence'
MQ_MODS_KEY = 'Modifications'
MQ_MZ_ERROR = 'Simple mass error [ppm]'
MQ_RT_KEY = 'Retention time'
MQ_SCAN_KEY = 'Scan number'
MQ_SCORE_KEY = 'Score'
MQ_SEQ_KEY = 'Sequence'
MQ_SOURCE_KEY = 'Raw file'
MQ_RELEVANT_COLS = [
    MQ_ACCESSION_KEY,
    MQ_CHARGE_KEY,
    MQ_DECOY_KEY,
    MQ_DELTA_SCORE_KEY,
    MQ_LEN_KEY,
    MQ_MASS_KEY,
    MQ_MISSED_CLEAVAGES_KEY,
    MQ_MOD_SEQ_KEY,
    MQ_MODS_KEY,
    MQ_MZ_ERROR,
    MQ_SCAN_KEY,
    MQ_SCORE_KEY,
    MQ_SOURCE_KEY,
    MQ_SEQ_KEY,
    MQ_RT_KEY,
]


def _create_mq_mods_df(mq_df):
    """ Helper function to create a DataFrame of the unique modifications found in
        MaxQuant search results.

    Parameters
    ----------
    mq_df : pd.DataFrame
        A DataFrame of MaxQuant search results.

    Returns
    -------
    mods_df : pd.DataFrame
        A small DataFrame containing the unique modifications found.
    """
    unique_mods = mq_df[MQ_MODS_KEY].apply(
        lambda x : x.split(',')
    ).explode().unique().to_list()

    unique_mods = list({x if x[1] != ' ' else x[2:] for x in unique_mods})
    unique_mods.remove('Unmodified')
    mods_df = pd.DataFrame({
        PTM_NAME_KEY: pd.Series(unique_mods),
        PTM_WEIGHT_KEY: pd.Series([KNOWN_PTM_WEIGHTS.get(mod, 0.0) for mod in unique_mods]),
    })
    mods_df = mods_df.sort_values(by=PTM_NAME_KEY)
    mods_df.reset_index(drop=True, inplace=True)
    mods_df[PTM_ID_KEY] = mods_df[PTM_NAME_KEY].apply(
        lambda x : ID_NUMBERS.get(x, 9)
    )
    mods_df = mods_df.sort_values(by=PTM_ID_KEY).reset_index(drop=True)
    mods_df[PTM_IS_VAR_KEY] = True
    return mods_df

def _create_ptm_seq_col(modified_seq, unqiue_mods):
    """ Helper function to take MaxQuant modified peptide and return
        the sequence representing the ptms.

    Parameters
    ----------
    modified_seq : str
        A modified peptide sequence in MaxQuant format.
    unqiue_mods : dict
        A dictionary mapping ptm names to integer ids.

    Returns
    -------
    ptm_seq : str
        The ptm sequence by ID.
    """
    ptm_seq = ''
    if '(' not in modified_seq:
        return None
    split_seq = modified_seq.split('_')
    if split_seq[0]:
        ptm_seq += f'{unqiue_mods[split_seq[0]]}.'
        main_seq = split_seq[1]
    elif split_seq[1].startswith('('):
        if split_seq[2].isupper():
            main_seq = split_seq[1]
            main_seq = main_seq[1:]
            end_mod = main_seq.index(')') + 1
            mod = main_seq[:end_mod]
            main_seq = main_seq[end_mod + 1:]
            ptm_seq += f'{str(unqiue_mods.get(mod, 9))}.'
        else:
            main_seq = split_seq[1]
            main_seq = main_seq[1:]
            end_mod = main_seq.index(')')
            mod = main_seq[:end_mod]
            main_seq = main_seq[end_mod + 1:]
            ptm_seq += f'{str(unqiue_mods.get(MQ_ABBREVIATIONS.get(mod, "unknown"), 9))}.'
    else:
        ptm_seq += '0.'
        main_seq = split_seq[1]
    while main_seq:
        assert main_seq[0] != '('
        if len(main_seq) == 1:
            ptm_seq += '0'
            break

        if main_seq[1] != '(':
            ptm_seq += '0'
            main_seq = main_seq[1:]
        else:
            if main_seq[2].isupper():
                main_seq = main_seq[2:]
                end_mod = main_seq.index(')') + 1
                mod = main_seq[:end_mod]
                main_seq = main_seq[end_mod + 1:]
                ptm_seq += str(unqiue_mods.get(mod, 9))
            else:
                main_seq = main_seq[2:]
                end_mod = main_seq.index(')')
                mod = main_seq[:end_mod]
                main_seq = main_seq[end_mod + 1:]
                ptm_seq += str(unqiue_mods.get(MQ_ABBREVIATIONS.get(mod, 'unknown'), 9))

    if split_seq[2]:
        ptm_seq += f'.{unqiue_mods.get(split_seq[2], 9)}'
    else:
        ptm_seq += '.0'

    return ptm_seq

def read_single_mq_data(mq_data):
    """ Function to read in MaxQuant search results from a single file.

    Parameters
    ----------
    df_loc : str
        A location of MaxQuant search results.

    Returns
    -------
    hits_df : pd.DataFrame
        A DataFrame of all search results properly formatted for inSPIRE.
    mods_dfs : pd.DataFrame
        A small DataFrame detailing the ptms found in the data.
    """
    mq_df = pl.read_csv(
        mq_data,
        separator='\t',
        columns=MQ_RELEVANT_COLS,
    )

    # Separate PTMs.
    mods_df = _create_mq_mods_df(mq_df)
    var_mod_dict = dict(zip(mods_df[PTM_NAME_KEY].tolist(), mods_df[PTM_ID_KEY].tolist()))
    mq_df = mq_df.with_columns(
        pl.col(MQ_MOD_SEQ_KEY).apply(
            lambda x : _create_ptm_seq_col(x, var_mod_dict)
        ).alias(PTM_SEQ_KEY)
    )

    # Rename to match inSPIRE naming scheme.
    mq_df = mq_df.rename({
        MQ_CHARGE_KEY: CHARGE_KEY,
        MQ_DELTA_SCORE_KEY: DELTA_SCORE_KEY,
        MQ_LEN_KEY: SEQ_LEN_KEY,
        MQ_MZ_ERROR: MASS_DIFF_KEY,
        MQ_RT_KEY: RT_KEY,
        MQ_SCAN_KEY: SCAN_KEY,
        MQ_SCORE_KEY: ENGINE_SCORE_KEY,
        MQ_SEQ_KEY: PEPTIDE_KEY,
        MQ_SOURCE_KEY: SOURCE_KEY,
        MQ_MISSED_CLEAVAGES_KEY: 'missedCleavages',
    })

    # Filter for Prosit, clean up accession data, and add label.
    mq_df = filter_for_prosit(mq_df)

    mq_df = mq_df.with_columns(
        pl.struct([MQ_ACCESSION_KEY, MQ_DECOY_KEY]).apply(
            lambda x : 'reverseSeq' if x[MQ_DECOY_KEY] == '+' else (
                x[MQ_ACCESSION_KEY] if x[MQ_ACCESSION_KEY] else 'unknown'
            ),
            skip_nulls=False,
        ).alias(ACCESSION_KEY)
    )

    mq_df = mq_df.with_columns(
        (pl.col(PEPTIDE_KEY).count()).over([SOURCE_KEY, SCAN_KEY]).alias('scanCounts')
    )
    mq_df = mq_df.with_columns(
        pl.col('scanCounts').gt(1).cast(int).alias('fromChimera'),
        skip_nulls=False,
    )

    mq_df = mq_df.with_columns(
        pl.struct([MQ_MASS_KEY, SEQ_LEN_KEY])
        .apply(
            lambda x : x[MQ_MASS_KEY]/x[SEQ_LEN_KEY],
            skip_nulls=False,
        ).alias('avgResidueMass')
    )
    mq_df = mq_df.with_columns(
        pl.col(MQ_DECOY_KEY)
        .apply(
            lambda x : -1 if x == '+' else 1,
            skip_nulls=False,
        ).alias(LABEL_KEY)
    )

    mq_df = mq_df.drop([
        MQ_ACCESSION_KEY,
        MQ_MOD_SEQ_KEY,
        MQ_DECOY_KEY,
        MQ_MASS_KEY,
        MQ_MODS_KEY,
        'scanCounts',
    ])

    return mq_df, mods_df

def read_mq_data(mq_data):
    """ Function to read in MaxQuant search results from one or more files.

    Parameters
    ----------
    peaks_data : str or list of str
        A single location of MaxQuant search results or a list of locations.

    Returns
    -------
    hits_df : pd.DataFrame
        A DataFrame of all search results properly formatted for inSPIRE.
    mods_dfs : pd.DataFrame
        A small DataFrame detailing the ptms found in the data.
    """
    if isinstance(mq_data, list):
        hits_dfs = []
        variable_mods_dfs = []

        for mq_file in mq_data:
            hits_df, mods_df = read_single_mq_data(mq_file)
            hits_dfs.append(hits_df)
            variable_mods_dfs.append(mods_df)

        # Combine DataFrames and validate that same PTMs are present.
        hits_df = pl.concat(hits_dfs)
        for i in range(len(variable_mods_dfs)-1):
            assert variable_mods_dfs[i].equals(variable_mods_dfs[i+1])
        return hits_df, variable_mods_dfs[0]

    return read_single_mq_data(mq_data)



""" Functions used across inSPIRE quantification pipeline.
"""
from math import log2

import pandas as pd
from plotly.colors import n_colors
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
import scipy
from scipy.stats import zscore
import seaborn as sns
from sklearn.decomposition import PCA

from inspire.constants import PLOT_AXIS_REQUIREMENTS
from inspire.logo_plot_utils import create_comparison_logo_plot


def plot_correlations(quant_file_name, code, config):
    """ Function to calculate the correlation between raw files and plot as heat map
        and write as html/svg.

    Parameters
    ----------
    quant_file_name : str
        Name of csv file containing quantification results.
    code : str
        Either "raw" or "norm" indicating if raw or normalised intensities are being compared.
    config : inspire.config.Config
        Config object for the whole experiment.
    """
    quant_df = pd.read_csv(
        f'{config.output_folder}/quant/{quant_file_name}.csv'
    )
    file_list = sorted([
        x for x in quant_df.columns if x.endswith(f'_{code}')
    ])

    if code == 'raw':
        for source in file_list:
            quant_df[source] = quant_df[source].apply(
                lambda intensity : log2(intensity) if intensity > 0 else None
            )

    correlations = {}
    for source in file_list:
        correlations[source] = []
        for source2 in file_list[::-1]:
            idp_key = source.replace(f'_{code}', '_idp')
            idp2_key = source2.replace(f'_{code}', '_idp')
            if code == 'raw':
                sub_quant_df = quant_df[
                    (quant_df[idp_key] > config.skyline_idp_cut_off) &
                    (quant_df[idp2_key] > config.skyline_idp_cut_off)
                ]

            sub_quant_df = quant_df[
                (quant_df[source].notna()) &
                (quant_df[source2].notna())
            ]

            correlations[source].append(
                round(
                    scipy.stats.pearsonr(
                        sub_quant_df[source],
                        sub_quant_df[source2],
                    )[0],
                    2,
                )
            )


    cor_df = pd.DataFrame(correlations)
    cor_df.index = file_list
    fig = go.Figure(
        go.Heatmap(
            z=cor_df.values,
            x=file_list,
            y=file_list[::-1],
            text=cor_df.values,
            texttemplate="%{text}",
            # textfont={"size":3},
            colorscale='RdBu_r',
            zmin=0,zmax=1,
        ),
    )

    fig.update_layout(
        width=500,
        height=300,
        font_size=8,
        font_family='Helvetica',
        font_color='black',
        margin={'r':25, 'l':25, 't':25, 'b':25},
    )

    fig.write_image(
        f'{config.output_folder}/img/{code}_correlation.svg', engine='kaleido'
    )


def plot_quant_clustermap(config):
    """ Function to plot clustermap of peptides and their binding affinities.

    Parameters
    ----------
    config : inspire.config.Config
        Config object for the whole experiment.
    """
    normed_df = pd.read_csv(f'{config.output_folder}/quant/normalised_quantification.csv')

    normed_df.index = normed_df['peptide']
    normed_df = normed_df[[col for col in normed_df.columns if col.endswith('_norm')]]

    normed_df = normed_df.rename(
        columns={col:col.strip('_norm') for col in normed_df.columns}
    )

    normed_df = normed_df.sort_values(by=list(normed_df.columns))
    normed_df = normed_df[normed_df.notnull().min(axis=1)]

    z_score_df = pd.DataFrame(
        zscore(normed_df, axis=1, nan_policy='omit'),
        index=normed_df.index,
        columns=normed_df.columns,
    )

    cluster_map = sns.clustermap(
        z_score_df, cmap="vlag", figsize=(10,30), cbar_pos=None, dendrogram_ratio=(0.1,0.03),
    )
    cluster_map.figure.savefig(f'{config.output_folder}/img/quant_clustermap.svg')

    z_score_df = z_score_df.transpose()
    pca = PCA(n_components=2).fit(z_score_df)

    pca_df = pd.DataFrame(
        pca.transform(z_score_df),
        columns=[
            f'Component 1 (Exp. Var. {round(pca.explained_variance_ratio_[0], 2)})',
            f'Component 2 (Exp. Var. {round(pca.explained_variance_ratio_[1], 2)})',
        ],
        index=z_score_df.index)
    pca_df /= pca_df.abs().max()
    pca_df['file'] = z_score_df.index
    pca_df['color'] = pca_df['file'].apply(
        lambda x : 'red' if x.startswith('inf') else (
            'blue' if x.startswith('control') else 'green'
        )
    )
    n_inf_files = pca_df[pca_df['color']=='red'].shape[0]
    n_cont_files = pca_df[pca_df['color']=='blue'].shape[0]
    n_unknown_files = pca_df[pca_df['color']=='green'].shape[0]

    colors1 = n_colors('rgb(0, 0, 0)', 'rgb(211, 211, 211)', n_cont_files, colortype='rgb')
    colors2 = n_colors('rgb(0, 100, 0)', 'rgb(144, 238, 144)', n_inf_files, colortype='rgb')
    colors3 = n_colors('rgb(0, 0, 128)', 'rgb(30, 144, 255)', n_unknown_files, colortype='rgb')

    fig = px.scatter(pca_df, x=pca_df.columns[0], y=pca_df.columns[1], color='file',
                     color_discrete_sequence=colors1+colors2+colors3)
    fig.update_traces(opacity=1)
    fig.update_traces(marker_line_color='black', marker_line_width=0.5)
    fig.update_traces(marker_size=8)
    fig.update_xaxes(PLOT_AXIS_REQUIREMENTS, range=[-2,2])
    fig.update_yaxes(PLOT_AXIS_REQUIREMENTS, range=[-2,2])
    fig.update_layout(
        barmode='group',
        font_family='Helvetica',
        font_color='black',
        width=600,
        height=500,
        title_x=0.5,
        plot_bgcolor='rgba(0,0,0,0)',
        showlegend=True,
    )
    fig.write_image(
        f'{config.output_folder}/img/quant_pca.svg', engine='kaleido'
    )


def plot_distros(quant_df, config, sources):
    """ Function to plot distribution of peptide intensities across raw files before and
        after normalisation.

    Parameters
    ----------
    quant_df : pd.DataFrame
        Quantifications of each peptide (rows) for each raw file (columns).
    config : inspire.config.Config
        Config object for the whole experiment.
    sources : list of str
        List of the raw files.
    """
    all_valids = []
    for source in sources:
        valid_quant_df = quant_df[
            (quant_df[f'{source}_clean'].apply(lambda a : a is not None)) &
            (quant_df[f'{source}_norm'].apply(lambda a : a is not None))
        ]
        valid_quant_df['File'] = source
        all_valids.append(valid_quant_df[[
            'File', f'{source}_clean', f'{source}_norm'
        ]].rename(columns={
            f'{source}_norm': 'Normalised Intensity',
            f'{source}_clean': 'Raw Intensity'
        }))
    total_df = pd.concat(all_valids)

    fig=go.Figure()
    fig.add_trace(go.Violin(
        x=total_df['File'],y=total_df['Raw Intensity'],
        fillcolor='#ACE1AF', meanline_visible=True,
        side='negative', name='Raw', points=False,
        line_color='black',
        line_width=0.5,
    ))
    fig.add_trace(go.Violin(
        x=total_df['File'],y=total_df['Normalised Intensity'],
        points=False,
        fillcolor='darkgreen',
        line_color='black',
        line_width=0.5,
        meanline_visible=True,side='positive', name='Normalised'))
    fig.update_traces(marker={'line': {'color': 'black', 'width': 0.5}})
    fig.add_hline(
        y=total_df['Normalised Intensity'].mean(),
        line_color='darkgreen',
        line_width=1.5,
        line_dash='dash',
    )
    fig.add_hline(
        y=total_df['Raw Intensity'].mean(),
        line_color='#ACE1AF',
        line_width=1.5,
        line_dash='dash',
    )
    fig.update_xaxes(PLOT_AXIS_REQUIREMENTS)
    fig.update_yaxes(PLOT_AXIS_REQUIREMENTS, range=[15,35])
    fig.update_layout(
        font_family='Helvetica',
        font_color='black',
        width=900,
        height=400,
        title_x=0.5,
        plot_bgcolor='rgba(0,0,0,0)',
        showlegend=True,
    )
    fig.write_image(
        f'{config.output_folder}/img/quant_distro.svg', engine='kaleido'
    )


def join_peptide(df_row, clusters):
    peptide = ''
    for cluster in clusters:
        if isinstance(df_row[cluster], str):
            peptide += df_row[cluster]
        else:
            peptide += 'X'
    return peptide

def create_quant_logoplots(de_df, output_folder, centers):
    """ Function to create quantitative logoplots based on normalised inspire dataframe
    """
    sorted_clusters = [x[1] for x in sorted(list(zip(
        centers, list(range(7))
    )))]
    create_violin_cluster_plot(de_df, sorted_clusters, output_folder)
    center_cluster = sorted_clusters.pop(3)
    center_cluster_c_term = de_df[
        de_df['cluster'] == center_cluster
    ]['peptide'].apply(lambda x : x[-1]*6)

    center_df = pd.DataFrame({
        'peptide' : center_cluster_c_term
    })
    cluster_df = pd.DataFrame({
        idx: de_df[
            de_df['cluster'] == idx
        ]['peptide'].apply(lambda x : x[-1]) for idx in sorted_clusters
    })
    cluster_df['peptide'] = cluster_df.apply(
        lambda x : join_peptide(x, sorted_clusters), axis=1
    )

    create_comparison_logo_plot(
        [cluster_df, center_df],
        ['C-terminus: Different Clusters', 'Center Cluster'],
        6,
        f'{output_folder}/img',
        'quant_logo_plots',
        data_index=[1,2,3,5,6,7]
    )

def create_violin_cluster_plot(de_df, sorted_clusters, output_folder):
    """ Create figure showing distribution of fold changes across the cluste
    """
    violin_fig = go.Figure()
    for idx, cluster in enumerate(sorted_clusters):
        cluster_df = de_df[de_df['cluster'] == cluster]
        violin_fig.add_trace(
            go.Violin(
                x=[idx+1]*cluster_df.shape[0], y=cluster_df['foldChange'],
                fillcolor='pink', line_color='black', line_width=0.5,
                points=False
            )
        )
     # Clean axes and layout
    violin_fig.update_xaxes(
        showline=False,
        showticklabels=True,
        title_text='Cluster',
        linecolor='black',
        linewidth=0.5,
        showgrid=False,
        ticks="outside",
        dtick=1,
    )
    violin_fig.add_hline(y=0, line_color='black', line_width=0.5, line_dash='dash')
    violin_fig.update_yaxes(
        showline=True,
        showticklabels=True,
        linewidth=0,
        linecolor='black',
        title_text='Log Fold Change',
        showgrid=False,
        ticks="outside",
    )
    violin_fig.update_layout(
        width=500,
        height=300,
        paper_bgcolor='rgba(256,256,256,256)',
        plot_bgcolor='rgba(256,256,256,256)',
        showlegend=False,
        font_color='black',
        font_family='Helvetica',
    )

    pio.write_image(
        violin_fig, f'{output_folder}/img/quant_cluster_violin.svg',
    )



""" Suite of function for preparing input to external tools, specifically
    Prosit and netMHCpan input.
"""

import os

import polars as pl

from inspire.constants import (
    CHARGE_KEY,
    ENDC_TEXT,
    MS2PIP_NAME_MAPPINGS,
    OKCYAN_TEXT,
    PEPTIDE_KEY,
    PTM_ID_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
    SEQ_LEN_KEY,
)
from inspire.input.search_results import generic_read_df
from inspire.utils import get_ox_flag

def create_prosit_mod_seq(sequence, modifications, ox_marker):
    """ Function to add any required oxidation flags to the peptide sequence.

    Parameters
    ----------
    sequence : str
        The original unmodified sequence.
    modifications : str or None
        The ptms associated with this sequence.
    ox_marker : int
        The ptm marker that indicates oxidation.

    Returns
    -------
    mod_seq : str
        The sequence with modifications added.
    """
    if (
        modifications is None or
        not modifications or
        modifications == 'nan' or
        not isinstance(modifications, str)
    ):
        return sequence

    ptms_list = modifications.split(".")
    mods_list = [int(mod) for mod in ptms_list[1]]
    if ox_marker in mods_list:
        mod_seq = ""
        previous_ind = 0
        for idx, mod in enumerate(mods_list):
            if mod == ox_marker:
                mod_seq += sequence[previous_ind:idx+1]
                mod_seq += '(ox)'
                previous_ind = idx+1
        mod_seq += sequence[previous_ind:]
        return mod_seq
    return sequence

def write_prosit_input_df(
        search_df,
        mods_df,
        config,
        collision_energy,
        filename,
        overwrite=True,
    ):
    """ Function to write prosit input sequences.

    Parameters
    ----------
    search_df : pd.DataFrame
        A DataFrame containing search results.
    mods_df : pd.DataFrame
        A DataFrame containing variable ptm data.
    output_folder : str
        The folder where output will be written.
    collision_energy : int
        The collision energy setting of the mass spectrometer.
    """
    # Create modified sequence.
    ox_flag = get_ox_flag(mods_df)
    search_df = search_df.with_columns(
        pl.struct([PEPTIDE_KEY, PTM_SEQ_KEY]).apply(
            lambda x : create_prosit_mod_seq(x[PEPTIDE_KEY], x[PTM_SEQ_KEY], ox_flag),
            skip_nulls=False,
        ).alias('modified_sequence')
    )

    # Write sequences in Prosit's input format.
    prosit_df = search_df.select('modified_sequence', CHARGE_KEY).rename(
        {CHARGE_KEY: 'precursor_charge'}
    )
    if isinstance(collision_energy, list):
        prosit_df = prosit_df.unique()
        prosit_df['collision_energy'] = [collision_energy for _ in range(prosit_df.shape[0])]
        prosit_df = prosit_df.explode('collision_energy')
        if overwrite:
            prosit_df.write_csv(
                f'{config.output_folder}/{filename}.csv',
            )
        else:
            with open(f'{config.output_folder}/{filename}.csv', mode='ab') as out_file:
                prosit_df.write_csv(
                    out_file,
                    has_header=False,
                )
    else:
        prosit_df = prosit_df.with_columns(
            pl.lit(collision_energy).alias('collision_energy')
        )
        prosit_df = prosit_df.unique()
        if overwrite:
            prosit_df.write_csv(
                f'{config.output_folder}/{filename}.csv',
            )
        else:
            with open(f'{config.output_folder}/{filename}.csv', mode='ab') as out_file:
                prosit_df.write_csv(
                    out_file,
                    has_header=False,
                )

def prepare_for_spectral_prediction(config):
    """ Function to prepare sequences for Prosit input.

    Parameters
    ----------
    configuration : inspire.config.Config
        The configuration settings for the pipeline.
    """
    target_df, mods_df = generic_read_df(config)

    if config.spectral_predictor == 'prosit':
        write_prosit_input_df(
            target_df,
            mods_df,
            config,
            config.collision_energy,
            'prositInput'
        )
        print(
            OKCYAN_TEXT +
            '\tFormatted Prosit input written.' +
            ENDC_TEXT
        )
    else:
        write_ms2pip_input_df(
            target_df,
            mods_df,
            config.output_folder,
            'ms2pipInput',
        )
        print(
            OKCYAN_TEXT +
            '\tFormatted MS2PIP input written.' +
            ENDC_TEXT
        )

def get_ms2pip_mods(ptm_seq, mod_id_mappings):
    """ Function to format PTMs for MS2PIP input.

    Parameters
    ----------
    ptm_seq : str or NaN
        The PTM sequence formatted by inSPIRE.
    mod_id_mappings : dict
        A dictionary mapping PTM IDs to the weight of the PTM.

    Return
    ------
    ms2pip_mods : str
        The PTMs observed formatted for MS2PIP input.
    """
    if not isinstance(ptm_seq, str):
        return '-'
    ptm_seq = ptm_seq.replace('.', '')
    mods = []
    for idx, char in enumerate(ptm_seq):
        if char != '0':
            mods.append(f'{idx}|{mod_id_mappings[int(char)]}')
    if mods:
        return '|'.join(mods)
    return '-'

def write_ms2pip_input_df(target_df, mods_df, output_folder, output_name):
    """ Function to write input sequences for ms2pip.

    Parameters
    ----------
    target_df : pl.DataFrame
        A DataFrame of PSMs for which we require MS2PIP predictions.
    mods_df : pl.DataFrame
        A small DataFrame detailing the PTMs in the data.
    output_folder : str
        The folder where all inSPIRE output is written.
    output_name : str
        The name used in the MS2PIP input file.
    delta_method : str
        The method used to calculate Prosit Delta values.
    """
    mods_df['ms2pipName'] = mods_df[PTM_NAME_KEY].apply(
        lambda x : MS2PIP_NAME_MAPPINGS[x]
    )
    mod_id_mappings = dict(zip(mods_df[PTM_ID_KEY].tolist(), mods_df['ms2pipName'].tolist()))

    target_df = target_df.with_row_count(name='spec_id')
    target_df = target_df.with_columns(
        pl.col('spec_id').apply(lambda x : f'peptide_{x}'),
        pl.col(PTM_SEQ_KEY).apply(
            lambda x : get_ms2pip_mods(x, mod_id_mappings)
        ).alias('modifications'),
    )

    ms2pip_input_df = target_df.select(['spec_id', 'modifications', 'peptide', 'charge'])
    ms2pip_input_df = ms2pip_input_df.unique(
        subset=['modifications', 'peptide', 'charge']
    )

    ms2pip_input_df.write_csv(
        f'{output_folder}/{output_name}.preprec', separator=' ',
    )


def prepare_for_mhcpan(config):
    """ Function to prepare sequences for NetMHCpan input.

    Parameters
    ----------
    configuration : inspire.config.Config
        The configuration settings for the pipeline.
    """
    if config.use_binding_affinity not in ('asValidation', 'asFeature'):
        return

    target_df, _ = generic_read_df(config)

    unique_pep_lens = target_df[SEQ_LEN_KEY].unique().to_list()

    if not os.path.exists(f'{config.output_folder}/mhcpan'):
        os.makedirs(f'{config.output_folder}/mhcpan')

    for length in unique_pep_lens:
        if length > config.ba_pred_limit:
            continue
        len_df = target_df.filter(pl.col(SEQ_LEN_KEY) == length)
        peptides = len_df.select(PEPTIDE_KEY).unique()
        peptides = peptides.with_row_count('id').with_columns(
            (pl.col('id')//10_000).alias('batch')
        )
        split_peptides = peptides.partition_by('batch')
        for idx, pep_group_df in enumerate(split_peptides):
            pep_group_df.select('peptide').write_csv(
                f'{config.output_folder}/mhcpan/inputLen{length}_{idx}.txt',
                has_header=False,
            )

    print(
        OKCYAN_TEXT +
        '\tFormatted NetMHCpan input written.' +
        ENDC_TEXT
    )



""" Functions for running Skyline quantification via a Docker image.
"""
import os
import shutil

import docker
import pandas as pd
import platform
import polars as pl

from inspire.constants import (
    ACCESSION_KEY,
    CHARGE_KEY,
    MOD_SEQ_KEY,
    PEPTIDE_KEY,
    FINAL_Q_VALUE_KEY,
    SOURCE_KEY,
)
from inspire.download import download_utils
from inspire.quant.utils import plot_correlations
from inspire.utils import is_control

SKYLINE_INTERMEDIATE_FILES = [
    'inspire_identifications.fasta',
    'inspire_identifications.ssl',
    'skyline_config.blib',
    'skyline_config.redundant.blib',
    'skyline_config.skyd',
    'skyline_config.slc',
    'skyline_report.csv',
    'skyline_report_template.skyr',
]
SKYLINE_AREA_KEY = 'TotalAreaMs1'
SKYLINE_BACKGROUND_KEY = 'TotalBackgroundMs1'

SKYLINE_INPUT_FILE_KEY = 'file'
SKYLINE_INPUT_SCAN_KEY = 'scan'
SKYLINE_INPUT_CHARGE_KEY = 'charge'
SKYLINE_INPUT_PEP_KEY = 'sequence'
SKYLINE_INPUT_SCORE_KEY = 'score'
SKYLINE_INPUT_KEYS = [
    SKYLINE_INPUT_FILE_KEY,
    SKYLINE_INPUT_SCAN_KEY,
    SKYLINE_INPUT_CHARGE_KEY,
    SKYLINE_INPUT_PEP_KEY,
    SKYLINE_INPUT_SCORE_KEY,
]

SKYLINE_OUTPUT_CHARGE_KEY = 'PrecursorCharge'
SKYLINE_OUTPUT_PEP_KEY = 'PeptideSequence'
SKYLINE_OUTPUT_FILE_KEY = 'FileName'
SKYLINE_OUTPUT_IDP_KEY = 'IsotopeDotProduct'


def quantify_identifications(config):
    """ Function to run Skyline quantification via the Docker image and format the output.

    Parameters
    ----------
    config : inspire.config.Config
        Config object for the whole experiment.
    """
    # Download necessary config files if necessary
    download_utils()

    # Prepare inputs
    create_skyline_input(config)
    copy_skyline_files_to_scans_dir(config)

    # Execute quantification
    execute_skyline(config)

    # Format output, add metadata.
    format_skyline_output(config)

    # Plot correlations between raw results.
    plot_correlations('quantified_per_file', 'raw', config)

def create_skyline_input(config):
    """ Function to format the identifications for Skyline.

    Parameters
    ----------
    config : inspire.config.Config
        Config object which controls the experiment.
    """
    output_df = pl.read_csv(f'{config.output_folder}/finalPsmAssignments.csv')
    output_df = output_df.with_columns(
        pl.struct([ACCESSION_KEY, PEPTIDE_KEY]).apply(
            lambda df_row : (
                f'>{df_row[ACCESSION_KEY]}_{df_row[PEPTIDE_KEY]}\n{df_row[PEPTIDE_KEY]}\n'
            )
        ).alias('fastaEntry')
    )

    # Write all peptides to the fasta file
    with open(
        f'{config.output_folder}/quant/inspire_identifications.fasta',
        mode='w',
        encoding='UTF-8',
    ) as out_fasta:
        output_df['fastaEntry'].apply(
            out_fasta.write
        )

    # Filter identifications
    quant_df = output_df.filter(
        pl.col(FINAL_Q_VALUE_KEY).lt(config.quantification_cut_off)
    )

    # Format for Skyline
    quant_df = quant_df.rename({
        MOD_SEQ_KEY: SKYLINE_INPUT_PEP_KEY,
    })
    quant_df = quant_df.with_columns(
        (pl.col(SOURCE_KEY) + '.raw').alias(SKYLINE_INPUT_FILE_KEY),
        pl.lit(0.001).alias(SKYLINE_INPUT_SCORE_KEY),
    )

    # Write identifications for Skyline.
    quant_df.select(
        SKYLINE_INPUT_KEYS
    ).write_csv(
        f'{config.output_folder}/quant/inspire_identifications.ssl',
        separator='\t',
    )


def copy_skyline_files_to_scans_dir(config):
    """ Function to copy the required Skyline files to the scans folder.

    Parameters
    ----------
    config : inspire.config.Config
        Config object which controls the experiment.
    """
    shutil.copyfile(
        f'{config.output_folder}/quant/inspire_identifications.fasta',
        f'{config.scans_folder}/inspire_identifications.fasta',
    )
    shutil.copyfile(
        f'{config.output_folder}/quant/inspire_identifications.ssl',
        f'{config.scans_folder}/inspire_identifications.ssl',
    )

    with open(
        config.skyline_config_file,
        mode='r',
        encoding='UTF-8',
    ) as skyline_config_file:
        skyline_config = skyline_config_file.read().format(
            precursor_ion_filter=f'{config.skyline_precursor_filter}',
        )
    with open(
        f'{config.scans_folder}/skyline_config.sky',
        mode='w',
        encoding='UTF-8',
    ) as params_file:
        params_file.write(skyline_config)

    shutil.copyfile(
        config.skyline_report_template,
        f'{config.scans_folder}/skyline_report_template.skyr',
    )


def execute_skyline(config):
    """ Function to execute Skyline via the docker image.

    Parameters
    ----------
    config : inspire.config.Config
        Config object for the whole experiment.
    scans_folder_abs_path : str
        The absolute path to the folder containing scan data (where inSPIRE
        has copied the other data required by Skyline).
    """
    scans_folder_abs_path = os.path.abspath(config.scans_folder)

    # Execute docker
    if platform.system() != 'Windows':
        client = docker.from_env()
        container = client.containers.run(
            'proteowizard/pwiz-skyline-i-agree-to-the-vendor-licenses',
            ' wine SkylineCmd --timestamp --dir=/data --in=skyline_config.sky ' +
            # ' --full-scan-rt-filter=ms2_ids --full-scan-rt-filter-tolerance=0.25 '
            ' --import-search-file=inspire_identifications.ssl ' +
            ' --import-fasta=inspire_identifications.fasta ' +
            ' --import-search-include-ambiguous ' +
            ' --report-conflict-resolution=overwrite ' +
            ' --report-add=skyline_report_template.skyr --report-name=myreport --report-invariant' +
            f' --report-file=skyline_report.csv',
            tty=True,
            stdin_open=True,
            auto_remove=True,
            volumes={scans_folder_abs_path: {'bind': '/data', 'mode': 'rw'}},
            detach=True
        )
        with open(f'{config.output_folder}/quant/skyline_log.txt', 'w') as log_file:
            # Stream logs and write them to the file
            for log in container.logs(stream=True):
                log_file.write(log.decode('utf-8'))
        container.stop()
    else:
        os.system(
            f'"{config.skyline_runner}"' +
            f' --timestamp --dir={scans_folder_abs_path} --in=skyline_config.sky ' +
            ' --import-search-file=inspire_identifications.ssl ' +
            ' --import-fasta=inspire_identifications.fasta ' +
            ' --report-conflict-resolution=overwrite ' +
            ' --import-search-include-ambiguous ' +
            ' --report-add=skyline_report_template.skyr --report-name=myreport --report-invariant' +
            f' --report-file=skyline_report.csv > {config.output_folder}/quant/skyline_log.txt',
        )

    # Copy output back
    if os.path.exists(f'{config.scans_folder}/skyline_report.csv'):
        shutil.copyfile(
            f'{config.scans_folder}/skyline_report.csv',
            f'{config.output_folder}/quant/skyline_report.csv',
        )
    else:
        raise RuntimeError(f'Skyline failed. Check log at: {config.output_folder}/quant/skyline_log.txt')

    # Remove intermediate files.
    for skyline_file in SKYLINE_INTERMEDIATE_FILES:
        if os.path.exists(f'{config.scans_folder}/{skyline_file}'):
            os.remove(f'{config.scans_folder}/{skyline_file}')


def format_skyline_output(config):
    """ Function to format skyline outputs - per file total area less background
        and IDP per peptide per file.

    Parameters
    ----------
    config : inspire.config.Config
        Config object which controls the experiment.
    """
    # Read in report and rename some columns
    quant_df = pd.read_csv(f'{config.output_folder}/quant/skyline_report.csv')
    quant_df = quant_df.rename(columns={
        SKYLINE_OUTPUT_PEP_KEY: PEPTIDE_KEY,
        SKYLINE_OUTPUT_CHARGE_KEY: CHARGE_KEY,
        SKYLINE_OUTPUT_FILE_KEY: SOURCE_KEY,
    })

    # Avoid IDP of 0 if area is 0.
    quant_df['IsotopeDotProduct'] = quant_df[[SKYLINE_AREA_KEY, SKYLINE_OUTPUT_IDP_KEY]].apply(
        lambda df_row : round(
            df_row[SKYLINE_OUTPUT_IDP_KEY], 3
        ) if df_row[SKYLINE_AREA_KEY] > 0 else 1.0,
        axis=1,
    )
    quant_df = quant_df.sort_values(by=SKYLINE_AREA_KEY, ascending=False)
    quant_df = quant_df.drop_duplicates(subset=[
        SOURCE_KEY, PEPTIDE_KEY
    ])

    # Filter rows where source not defined, remove the ".raw" suffix, and collect unique sources.
    quant_df = quant_df[
        quant_df[SOURCE_KEY].apply(
            lambda x : isinstance(x, str)
        )
    ]
    quant_df[SOURCE_KEY] = quant_df[SOURCE_KEY].apply(lambda x : x[:-4])
    sources = quant_df[SOURCE_KEY].unique().tolist()

    # Calculate ratio of area to background and get area less background (min of 0).
    quant_df['AreaBackgroundRatio'] = quant_df[SKYLINE_AREA_KEY]/(
        quant_df[SKYLINE_BACKGROUND_KEY]+quant_df[SKYLINE_AREA_KEY]
    )
    quant_df['AreaLessBackground'] = (
        quant_df[SKYLINE_AREA_KEY] - quant_df[SKYLINE_BACKGROUND_KEY]
    )
    quant_df['AreaLessBackground'] = quant_df['AreaLessBackground'].apply(
        lambda x : x if x > 0 else 0
    )

    # Pivot peptide table and add IDP and add area/background ratio inforamtion.
    quant_area_df = pivot_quant_data(quant_df, 'AreaLessBackground', '_raw')
    quant_idp_df = pivot_quant_data(quant_df, 'IsotopeDotProduct', '_idp')
    quant_ratio_df = pivot_quant_data(quant_df, 'AreaBackgroundRatio', '_ratio')
    pivot_quant_df = pd.merge(
        quant_area_df,
        quant_idp_df,
        how='inner',
        on=PEPTIDE_KEY,
    )
    pivot_quant_df = pd.merge(
        pivot_quant_df,
        quant_ratio_df,
        how='inner',
        on=PEPTIDE_KEY,
    )

    # If not quantified fill with idp 0.
    for source in sources:
        pivot_quant_df[f'{source}_idp'] = pivot_quant_df[f'{source}_idp'].fillna(0.0)
        pivot_quant_df[f'{source}_idp'] = pivot_quant_df[f'{source}_idp'].replace(0.0, 1.0)

    # Add identification data, variation in technical replicates and rename columns
    pivot_quant_df = add_id_data(pivot_quant_df, sources, config)
    pivot_quant_df = rename_raw_file_columns(pivot_quant_df, config, sources)

    # Ensure first two columns are peptide and accession:
    quant_cols = [col for col in pivot_quant_df.columns if col not in (PEPTIDE_KEY, ACCESSION_KEY)]
    pivot_quant_df = pivot_quant_df[[PEPTIDE_KEY, ACCESSION_KEY] + quant_cols]

    # Write output
    pivot_quant_df.to_csv(
        f'{config.output_folder}/quant/quantified_per_file.csv',
        index=False,
    )


def pivot_quant_data(quant_df, value_key, suffix):
    """ Function to pivot quantification data on the peptide column.

    Parameters
    ----------
    quant_df : pd.DataFrame
        DataFrame of quantified peptides across multiple raw files.
    value_key : str
        The key which
    suffix : str
        The suffix which should be added to the value columns.

    Return
    ------
    pivot_quant_df : pd.DataFrame
        Quantification data pivoted by peptide index.
    """
    pivot_quant_df = quant_df.pivot(
        index=[PEPTIDE_KEY],
        columns=SOURCE_KEY,
        values=value_key,
    )
    pivot_quant_df = pivot_quant_df.reset_index()
    sources = [x for x in pivot_quant_df.columns if x != PEPTIDE_KEY]

    pivot_quant_df = pivot_quant_df.rename(columns={
        source: f'{source}{suffix}' for source in sources
    })

    return pivot_quant_df


def add_id_data(pivot_quant_df, sources, config):
    """ Function to add 1/0 flags for each peptide of whether the peptide was identified
        in each raw file of the experiment.

    Parameters
    ----------
    pivot_quant_df : pd.DataFrame
        A DataFrame pivoted on peptide with quantification across each file.
    sources : list of str
        A list of the raw files in the quantification data.
    config : inspire.config.Config
        Config object which controls the experiment.

    Returns
    -------
    pivot_quant_df : pd.DataFrame
        The input DataFrame with identification columns added.
    """
    # Read and filter assignments by q-value and remove duplicates within a raw file.
    identification_df = pd.read_csv(f'{config.output_folder}/finalPsmAssignments.csv')
    identification_df = identification_df[
        identification_df[FINAL_Q_VALUE_KEY] < config.quantification_cut_off
    ]
    identification_df = identification_df.drop_duplicates(
        subset=[PEPTIDE_KEY, ACCESSION_KEY, SOURCE_KEY]
    )

    # Groupby peptide and accession and create columns for identified in each file.
    identification_df = identification_df.groupby(
        [PEPTIDE_KEY, ACCESSION_KEY],
        as_index=False,
    )[SOURCE_KEY].apply(list)
    for source in sources:
        identification_df[f'{source}_identified'] = identification_df[
            SOURCE_KEY
        ].apply(lambda x, src=source : int(src in x))
    identification_df = identification_df.drop(SOURCE_KEY, axis=1)

    # Merge identification data with quantification data.
    pivot_quant_df = pd.merge(
        pivot_quant_df,
        identification_df,
        how='inner',
        on='peptide',
    )
    return pivot_quant_df


def rename_raw_file_columns(pivot_quant_df, config, sources):
    """ Function to rename columns to short names and write name mappings.

    Parameters
    ----------
    pivot_quant_df : pd.DataFrame
        DataFrame of quantification data.
    config : inspire.config.Config
        Config object which controls the experiment.
    sources : list of str
        A list of the raw files in the quantification data.

    Returns
    -------
    pivot_quant_df : pd.DataFrame
        DataFrame of quantification data with updated column names.
    """
    if config.control_flags is not None:
        control_sources = sorted(
            [source for source in sources if is_control(source, config.control_flags)]
        )
        inf_sources = sorted(
            [source for source in sources if not is_control(source, config.control_flags)]
        )
        for source_idx, source in enumerate(control_sources):
            pivot_quant_df = pivot_quant_df.rename(columns={
                f'{source}_raw': f'controlFile{source_idx+1}_raw',
                f'{source}_idp': f'controlFile{source_idx+1}_idp',
                f'{source}_ratio': f'controlFile{source_idx+1}_ratio',
                f'{source}_identified': f'controlFile{source_idx+1}_identified',
            })
        for source_idx, source in enumerate(inf_sources):
            pivot_quant_df = pivot_quant_df.rename(columns={
                f'{source}_raw': f'infectionFile{source_idx+1}_raw',
                f'{source}_idp': f'infectionFile{source_idx+1}_idp',
                f'{source}_ratio': f'infectionFile{source_idx+1}_ratio',
                f'{source}_identified': f'infectionFile{source_idx+1}_identified',
            })
        file_renaming_df = pd.DataFrame({
            'source': control_sources + inf_sources,
            'renamed': [
                f'controlFile{idx+1}' for idx in range(len(control_sources))
            ] + [
                f'infectionFile{idx+1}' for idx in range(len(inf_sources))
            ],
            'status': ['control']*len(control_sources) + ['infected']*len(inf_sources),
        })
    else:
        sources = sorted(sources)
        for source_idx, source in enumerate(sources):
            pivot_quant_df = pivot_quant_df.rename(columns={
                f'{source}_raw': f'file{source_idx+1}_raw',
                f'{source}_idp': f'file{source_idx+1}_idp',
                f'{source}_ratio': f'file{source_idx+1}_ratio',
                f'{source}_identified': f'file{source_idx+1}_identified',
            })
        file_renaming_df = pd.DataFrame({
            'source': sources,
            'renamed': [
                f'file{idx+1}' for idx in range(len(sources))
            ],
        })

    file_renaming_df.to_csv(f'{config.output_folder}/quant/metadata.csv', index=False)

    return pivot_quant_df



""" Functions for applying the prosit-delta predictor.
"""
from math import acos, pi
from copy import deepcopy
import numpy as np

from inspire.constants import (
    BLOSUM6_1_VALUES,
    CHARGE_KEY,
    ION_OFFSET,
    KNOWN_PTM_WEIGHTS,
    PEPTIDE_KEY,
    PTM_SEQ_KEY,
    RESIDUE_WEIGHTS,
    SPECTRAL_ANGLE_KEY,
)
from inspire.mz_match import match_mz

DELTA_PRO_FEATURE_SET = (
    'spectralAngle',
    'blosumC',
    'blosumN',
    'nTermDist',
    'cTermDist',
    'charge',
    'cNeighbourBlosum',
    'nNeighbourBlosum',
    'yErrsAtLoc',
    'bErrsAtLoc',
    'collisionEnergy',
    'bPrositIntesAtC',
    'yPrositIntesAtN',
    'yPrositIntesAtLoc',
    'bPrositIntesAtLoc',
    'yMatchedIntesAtN',
    'bMatchedIntesAtC',
    'yMatchedIntesAtLoc',
    'bMatchedIntesAtLoc',
    'cOxidation',
    'nOxidation',
    'flipYNewIntensity',
    'flipBNewIntensity',
    'matchedCoverage',
)

SPECTRAL_ANGLE_INDEX = DELTA_PRO_FEATURE_SET.index('spectralAngle')
BLOSUM_N_INDEX = DELTA_PRO_FEATURE_SET.index('blosumN')
BLOSUM_C_INDEX = DELTA_PRO_FEATURE_SET.index('blosumC')
CHARGE_INDEX = DELTA_PRO_FEATURE_SET.index('charge')
MATCHED_COV_INDEX = DELTA_PRO_FEATURE_SET.index('matchedCoverage')
CE_INDEX = DELTA_PRO_FEATURE_SET.index('collisionEnergy')
C_NEIGHB_INDEX = DELTA_PRO_FEATURE_SET.index('cNeighbourBlosum')
N_NEIGHB_INDEX = DELTA_PRO_FEATURE_SET.index('nNeighbourBlosum')
N_TERM_INDEX = DELTA_PRO_FEATURE_SET.index('nTermDist')
C_TERM_INDEX = DELTA_PRO_FEATURE_SET.index('cTermDist')
Y_NEW_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('flipYNewIntensity')
B_NEW_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('flipBNewIntensity')
Y_ERR_INDEX = DELTA_PRO_FEATURE_SET.index('yErrsAtLoc')
B_ERR_INDEX = DELTA_PRO_FEATURE_SET.index('bErrsAtLoc')
B_PROSIT_C_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('bPrositIntesAtC')
Y_PROSIT_N_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('yPrositIntesAtN')
Y_PROSIT_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('yPrositIntesAtLoc')
B_PROSIT_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('bPrositIntesAtLoc')
Y_MATCHED_N_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('yMatchedIntesAtN')
B_MATCHED_C_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('bMatchedIntesAtC')
Y_MATCHED_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('yMatchedIntesAtLoc')
B_MATCHED_INTE_INDEX = DELTA_PRO_FEATURE_SET.index('bMatchedIntesAtLoc')
N_OX_INDEX = DELTA_PRO_FEATURE_SET.index('nOxidation')
C_OX_INDEX = DELTA_PRO_FEATURE_SET.index('cOxidation')

DELTA_PRO_RESIDUE_WEIGHTS = deepcopy(RESIDUE_WEIGHTS)
DELTA_PRO_RESIDUE_WEIGHTS['C'] = 103.009185 + 57.021464
DELTA_PRO_RESIDUE_WEIGHTS['m'] = DELTA_PRO_RESIDUE_WEIGHTS['M'] + KNOWN_PTM_WEIGHTS['Oxidation (M)']
SIGNIFICANCE_THRESHOLD = -0.05

def check_oxidation(pep_len, ptm_seq, idx, ox_flag):
    """ Helper function to check if resiude is oxidised.
    """
    if not isinstance(ptm_seq, str) or ox_flag not in ptm_seq or idx < 0 or idx > pep_len:
        return 0
    if ptm_seq[idx+2] == ox_flag:
        return 1
    return 0

def get_mass_diff(peptide, ptm_seq, idx, ox_flag):
    """ Function to find the difference between two residue masses.
    """
    if not isinstance(ptm_seq, str):
        return abs(
            DELTA_PRO_RESIDUE_WEIGHTS[peptide[idx]] - DELTA_PRO_RESIDUE_WEIGHTS[peptide[idx-1]]
        )

    c_wt = DELTA_PRO_RESIDUE_WEIGHTS[peptide[idx]]
    n_wt = DELTA_PRO_RESIDUE_WEIGHTS[peptide[idx-1]]
    if ptm_seq[idx+2] == ox_flag:
        c_wt += KNOWN_PTM_WEIGHTS['Oxidation (M)']
    if ptm_seq[idx+1] == ox_flag:
        n_wt += KNOWN_PTM_WEIGHTS['Oxidation (M)']
    return abs(c_wt - n_wt)

def get_intes_at_loc(pep_len, matched_intensities, prosit_ions, loc, letter):
    """ Function to calculate the sum of all intensities at a given location.

    Parameters
    ----------
    pep_len : int
        The length of the peptide.
    prosit_ions : dict
        The prosit predicted spectrum.
    loc : int
        The location at which to get intensities.
    letter : str
        The type of ion considered.

    Returns
    -------
    sum_inte : float
        The sum of intensities at the location.
    """
    sum_inte = 0.0
    if loc in (0, pep_len):
        return sum_inte

    if letter == 'y':
        loc = pep_len - loc

    for charge in ['', '^2', '^3']:
        for loss in ['', '-NH3', '-H2O']:
            full_code = letter + str(loc) + charge + loss
            inte_val = matched_intensities[prosit_ions == full_code]
            if inte_val.size:
                sum_inte += inte_val[0]
    return sum_inte

def compute_single_mz(sequence, modifications, ion_type, ion_idx, ptm_id_weights):
    """ Function to compute the molecular weights of potential fragments
        generated from a peptide (y & b ions, charges 1,2, or 3, and H2O
        or O2 losses).

    Parameters
    ----------
    sequence : str
        The peptide sequence for which we require molecular weights.
    modifications : str
        A string of the ptms for the sequence which will alter
        the potential mzs.
    reverse : bool
        Whether we are getting fragment mzs in the forward direction
        (eg for b ions), or backward direction (eg. for y ions).
    ptm_id_weights : dict
        Mapping of ptm ids to their molecular weights.

    Returns
    -------
    mzs : np.array of floats
        An array of all the possible mzs that coule be observed in
        the MS2 spectrum of a sequence.
    """
    if (
        modifications and
        isinstance(modifications, str) and
        modifications != 'nan'
        and modifications != 'None'
    ):
        ptms_list = modifications.split(".")
        mods_list = [int(mod) for mod in ptms_list[1]]

        if ion_type == 'y':
            ptm_start = int(ptms_list[2])
            mods_list = mods_list[::-1]
        else:
            ptm_start = int(ptms_list[0])
    else:
        mods_list = None
        ptm_start = 0.0

    if ion_type == 'y':
        sequence = sequence[::-1]

    tracking_mw = ptm_id_weights[ptm_start]

    for idx in range(ion_idx):
        tracking_mw += RESIDUE_WEIGHTS[sequence[idx]]
        if mods_list is not None and mods_list[idx]:
            tracking_mw += ptm_id_weights[mods_list[idx]]

    return tracking_mw + ION_OFFSET[ion_type]

def calculate_sa_from_dict(predicted, true):
    """ Function to calculate the spectral angle when the predicted and true spectrum
        are stored as dictionaries.

    Parameters
    ----------
    predicted : dict
        A dictionary of the prosit predicted spectrum.
    true : dict
        A dictionary of the experimental ions matched to the predicted spectrum.

    Returns
    -------
    spectral_angle : float
        The spectral angle between spectra.
    """
    true_l2_norm = np.linalg.norm(np.array(list(true.values())), ord=2)
    pred_l2_norm = np.linalg.norm(np.array(list(predicted.values())), ord=2)

    if true_l2_norm == 0 and pred_l2_norm == 0:
        return 0.0
    if true_l2_norm == 0 and pred_l2_norm == 0:
        return 1.0

    if pred_l2_norm > 0.0:
        for ion in predicted:
            predicted[ion] /= pred_l2_norm
    if true_l2_norm > 0.0:
        for ion in true:
            true[ion] /= true_l2_norm

    product = 0.0
    for ion in predicted:
        product += (predicted[ion] * true.get(ion, 0.0))

    product = max(min(product, 1.0), 0.0)
    spectral_distance = 2*acos(product)/pi
    return 1.0 - spectral_distance

def convert_ptm_seq(ptm_seq, flip_idx):
    """ Function to adjust a PTM sequence to match permutation of the sequence.

    Parameters
    ----------
    ptm_seq : str or None
        The the original PTM sequence (stored as string of digits)
    flip_idx : int
        The index which is to be permuted.

    Returns
    -------
    ptm_seq : str or None
        The adjusted PTM sequence.
    """
    if not isinstance(ptm_seq, str):
        return ptm_seq

    ptm_seq = (
        ptm_seq[:flip_idx+1] + ptm_seq[flip_idx+2] +
        ptm_seq[flip_idx+1] + ptm_seq[flip_idx+3:]
    )
    return ptm_seq

def calculate_delta_features(results, prot_deltas):
    """ Function to calculate the Prosit delta features.

    Parameters
    ----------
    df_row : pd.Series
        A single row of the PSM dataframe.
    prot_deltas : np.array
        The Prosit-delta values at each point in the sequence.

    Returns
    -------
    df_row : pd.Series
        The input row with Prosit-delta features added.
    """
    if prot_deltas.size != 0:
        results['nDeltasAboveThreshold'] = len(
            prot_deltas[prot_deltas > -0.1]
        )/len(prot_deltas)
        results['nDeltasAboveThresholdA'] = len(
            prot_deltas[prot_deltas > -0.05]
        )/len(prot_deltas)
        results['nDeltasAboveZero'] = len(
            prot_deltas[prot_deltas > 0.0]
        )/len(prot_deltas)
        delta_q1 = np.quantile(prot_deltas, 0.25)
        delta_q3 = np.quantile(prot_deltas, 0.75)
        results['prositDeltaMedian'] = float(np.quantile(prot_deltas, 0.5))
        results['prositDeltaQuartile1'] = float(delta_q1)
        results['prositDeltaQuartile3'] = float(delta_q3)
        results['minPrositDelta'] = float(np.min(prot_deltas))
        results['maxPrositDelta'] = float(np.max(prot_deltas))
    else:
        results['nDeltasAboveThreshold'] = -1.0
        results['nDeltasAboveThresholdA'] = -1.0
        results['nDeltasAboveZero'] = -1.0
        results['prositDeltaMedian'] = -1.0
        results['prositDeltaQuartile1'] = -1.0
        results['prositDeltaQuartile3'] = -1.0
        results['maxPrositDelta'] = -1.0
        results['minPrositDelta'] = -1.0

    return results

def compute_pd_single_mw(sequence, ion_type, ion_idx):
    """ Function to compute the mw of a single ion.

    Parameters
    ----------
    sequence : str
        The peptide sequence.
    ion_type : str
        y or b.
    ion_idx : int
        The index of ion.

    Returns
    -------
    mw : float
        The mw of the theoretical ion.
    """
    if ion_type == 'y':
        sequence = sequence[::-1]

    tracking_mw = 0.0

    for idx in range(ion_idx):
        tracking_mw += DELTA_PRO_RESIDUE_WEIGHTS[sequence[idx]]

    return tracking_mw + ION_OFFSET[ion_type]


def get_new_inte(peptide, flip_ind, charge, mzs, intes, mz_accuracy, ion_series):
    """ Function to get the potential new intensity.
    """
    if ion_series == 'b':
        base_mass = compute_pd_single_mw(peptide, 'b', flip_ind-1)
        base_mass += DELTA_PRO_RESIDUE_WEIGHTS[peptide[flip_ind]]
    else:
        base_mass = compute_pd_single_mw(peptide, 'y', len(peptide)-(flip_ind+1))
        base_mass += DELTA_PRO_RESIDUE_WEIGHTS[peptide[flip_ind-1]]
    match_inte = 0.0
    for frag_z in range(1, min(charge, 4)):
        mz_match, match_idx = match_mz(base_mass, frag_z, mzs, loss=0.0)
        if abs(mz_match) < mz_accuracy:
            match_inte += intes[match_idx]
    return match_inte

def get_err_at_loc(pep_len, matched_ions, prosit_ions, loc, letter):
    """ Function to get the error at a point in a spectrum between Prosit predicted and
        normalised experimental intensitiy.

    Parameters
    ----------
    pep_len : int
        Length of the peptide.
    matched_ions : dict
        Dictionary of the experimental intensities matched to Prosit ions.
    prosit_ions : dict
        Dictionary of the prosit predicted intensities.
    loc : int
        The location of interest.
    letter : str
        y or b

    Returns
    -------
    sum_error : float
        The sum of all errors at that location.
    """
    sum_err = 0
    if letter == 'y':
        loc = pep_len - loc
    for charge in ['', '^2', '^3']:
        code = letter + str(loc) + charge
        sum_err += abs(matched_ions.get(code, 0.0) - prosit_ions.get(code, 0.0))

    return sum_err

def get_deltas(
        results,
        df_row,
        mzs,
        intes,
        prosit_preds,
        prosit_ions,
        matched_intes,
        prosit_intes,
        reg_model,
        ox_flag,
        mz_accuracy,
        matched_dict,
    ):
    """ Function to get all predicted prosit-deltas.

    Parameters
    ----------
    df_row : pd.Series
        A row of a DataFrame containing spectral data.
    prosit_ions : dict
        The prosit predicted spectrum.
    reg_model : xgb.XGBRegressor
        The trained prosit-delta model.

    Returns
    -------
    df_row : pd.Series
        The updated row containing prosit-delta features.
    """
    peptide = df_row[PEPTIDE_KEY]
    charge = df_row[CHARGE_KEY]
    mod_seq = df_row['modified_sequence'].replace('M(ox)', 'm')

    true_l2_norm = np.linalg.norm(np.array(list(matched_dict.values())), ord=2)
    normed_matched_dict = {k:v/true_l2_norm for k,v in matched_dict.items()}
    pep_len = len(peptide)
    flip_inds = np.array([
        idx+1 for idx in range(
            len(df_row[PEPTIDE_KEY])-1
        ) if df_row[PEPTIDE_KEY][idx] != df_row[PEPTIDE_KEY][idx+1]
    ])

    if len(flip_inds):
        input_feats = np.zeros(
            shape=(len(flip_inds), len(DELTA_PRO_FEATURE_SET)), dtype=np.float64
        )
        input_feats[:, SPECTRAL_ANGLE_INDEX] = results[SPECTRAL_ANGLE_KEY]
        input_feats[:, BLOSUM_N_INDEX] = np.array([
            BLOSUM6_1_VALUES[peptide[idx-1]] for idx in flip_inds
        ])
        input_feats[:, BLOSUM_C_INDEX] = np.array([
            BLOSUM6_1_VALUES[peptide[idx]] for idx in flip_inds
        ])
        input_feats[:, CHARGE_INDEX] = df_row['charge']
        input_feats[:, MATCHED_COV_INDEX] = results['matchedCoverage']
        input_feats[:, CE_INDEX] = df_row['collisionEnergy']
        input_feats[:, C_NEIGHB_INDEX] = np.array(
            [-5.0 if idx > pep_len-3 else BLOSUM6_1_VALUES[peptide[idx+1]] for idx in flip_inds]
        )
        input_feats[:, N_NEIGHB_INDEX] = np.array(
            [-5.0 if idx < 2 else BLOSUM6_1_VALUES[peptide[idx-2]] for idx in flip_inds]
        )
        input_feats[:, N_TERM_INDEX] = np.array(flip_inds)
        input_feats[:, C_TERM_INDEX] = np.array([
            [pep_len-idx for idx in flip_inds]
        ])

        input_feats[:, Y_NEW_INTE_INDEX] = np.array(
            [
                get_new_inte(
                    mod_seq, flip_ind, charge, mzs, intes, mz_accuracy, 'y'
                ) for flip_ind in flip_inds
            ]
        )

        input_feats[:, B_NEW_INTE_INDEX] = np.array(
            [
                get_new_inte(
                    mod_seq, flip_ind, charge, mzs, intes, mz_accuracy, 'b'
                ) for flip_ind in flip_inds
            ]
        )
        input_feats[:, Y_ERR_INDEX] = np.array(
            [
                get_err_at_loc(
                    pep_len, normed_matched_dict, prosit_preds, flip_ind, 'y'
                ) for flip_ind in flip_inds
            ]
        )

        input_feats[:, B_ERR_INDEX] = np.array(
            [
                get_err_at_loc(
                    pep_len, normed_matched_dict, prosit_preds, flip_ind, 'b'
                ) for flip_ind in flip_inds
            ]
        )
        input_feats[:, B_PROSIT_C_INTE_INDEX] = np.array(
            [
                get_intes_at_loc(
                    pep_len,
                    prosit_intes,
                    prosit_ions,
                    flip_idx+1,
                    'b',
                ) for flip_idx in flip_inds
            ]
        )
        input_feats[:, Y_PROSIT_N_INTE_INDEX] = np.array(
            [
                get_intes_at_loc(
                    pep_len,
                    prosit_intes,
                    prosit_ions,
                    flip_idx-1,
                    'y',
                ) for flip_idx in flip_inds
            ]
        )

        input_feats[:, Y_PROSIT_INTE_INDEX] = np.array(
            [
                get_intes_at_loc(
                    pep_len,
                    prosit_intes,
                    prosit_ions,
                    flip_idx,
                    'y',
                ) for flip_idx in flip_inds
            ]
        )
        input_feats[:, B_PROSIT_INTE_INDEX] = np.array(
            [
                get_intes_at_loc(
                    pep_len,
                    prosit_intes,
                    prosit_ions,
                    flip_idx,
                    'b',
                ) for flip_idx in flip_inds
            ]
        )

        input_feats[:, Y_MATCHED_N_INTE_INDEX] = np.array(
            [
                get_intes_at_loc(
                    pep_len,
                    matched_intes,
                    prosit_ions,
                    flip_idx-1,
                    'y',
                ) for flip_idx in flip_inds
            ]
        )
        input_feats[:, B_MATCHED_C_INTE_INDEX] = np.array(
            [
                get_intes_at_loc(
                    pep_len,
                    matched_intes,
                    prosit_ions,
                    flip_idx+1,
                    'b'
                ) for flip_idx in flip_inds
            ]
        )
        input_feats[:, Y_MATCHED_INTE_INDEX] = np.array(
            [
                get_intes_at_loc(
                    pep_len,
                    matched_intes,
                    prosit_ions,
                    flip_idx,
                    'y'
                ) for flip_idx in flip_inds
            ]
        )
        input_feats[:, B_MATCHED_INTE_INDEX] = np.array(
            [
                get_intes_at_loc(
                    pep_len,
                    matched_intes,
                    prosit_ions,
                    flip_idx,
                    'b'
                ) for flip_idx in flip_inds
            ]
        )

        input_feats[:, C_OX_INDEX] = np.array(
            [
                check_oxidation(
                    pep_len,
                    df_row[PTM_SEQ_KEY],
                    flip_idx,
                    ox_flag,
                ) for flip_idx in flip_inds
            ]
        )
        input_feats[:, N_OX_INDEX] = np.array(
            [
                check_oxidation(
                    pep_len,
                    df_row[PTM_SEQ_KEY],
                    flip_idx-1,
                    ox_flag,
                ) for flip_idx in flip_inds
            ]
        )

        prot_deltas = reg_model.predict(input_feats)
    else:
        prot_deltas = np.array([])

    return calculate_delta_features(results, prot_deltas)



""" Functions for reading in scans results in mgf format.
"""
import re

import numpy as np
import polars as pl
from pyteomics import mgf

from inspire.constants import (
    CHARGE_KEY,
    INTENSITIES_KEY,
    MZS_KEY,
    RT_KEY,
    SCAN_KEY,
    SOURCE_KEY,
)

def process_mgf_file(
        mgf_filename,
        scan_ids,
        scan_file_format,
        source_list,
        combined_source_file=False,
        with_charge=False,
        with_retention_time=False,
        with_ms1=False,
    ):
    """ Function to process an mgf file to find matches with scan IDs.

    Parameters
    ----------
    mgf_filename : str
        The mgf file from which we are reading.
    scan_ids : set of int
        A set of the scan IDs we require.
    scan_file_format : str
        The format of the file used.
    source_list : list of str
        A list of source names.

    Returns
    -------
    scans_df : pd.DataFrame
        A DataFrame of scan results.
    """
    matched_scan_ids = []
    matched_intensities = []
    matched_mzs = []
    sources = []
    filename = mgf_filename.split('/')[-1]
    if with_charge:
        charge_list = []
    if with_retention_time:
        rt_list = []
    if with_ms1:
        ms1_intes = []

    with mgf.read(mgf_filename) as reader:
        for spectrum in reader:
            if scan_file_format is None or scan_file_format == 'distiller':
                if 'scans' in spectrum['params']:
                    scan_id = int(spectrum['params']['scans'])
                else:
                    regex_match = re.match(
                        r'(\d+)(.*?)',
                        spectrum['params']['title'].split('scan=')[-1]
                    )
                    scan_id = int(regex_match.group(1))
                if combined_source_file:
                    source = spectrum['params']['title'].split('File:"')[-1].split('.raw"')[0]
                else:
                    source = filename[:-4]
            else:
                scan_id = int(spectrum['params']['title'].split(' Scan ')[-1].split(' (rt')[0])
                source = source_list[
                    int(spectrum['params']['title'].split(' from file [')[-1].strip(']'))
                ]

            if scan_ids is None or scan_id in scan_ids:
                sources.append(source)
                matched_scan_ids.append(scan_id)
                matched_intensities.append(np.array(list(spectrum['intensity array'])))
                matched_mzs.append(np.array(list(spectrum['m/z array'])))
                if with_charge:
                    charge_list.append(int(spectrum['params']['charge'][0]))
                if with_retention_time:
                    rt_list.append(float(spectrum['params']['rtinseconds']))
                if with_ms1:
                    try:
                        ms1_intes.append(float(spectrum['params']['pepmass'][1]))
                    except Exception: # pylint: disable=bare-except
                        ms1_intes.append(-1.0)

    mgf_data = {
        SOURCE_KEY: pl.Series(sources),
        SCAN_KEY: pl.Series(matched_scan_ids),
        INTENSITIES_KEY: pl.Series(matched_intensities),
        MZS_KEY: pl.Series(matched_mzs)
    }

    if with_charge:
        mgf_data[CHARGE_KEY] = pl.Series(charge_list)
    if with_retention_time:
        mgf_data[RT_KEY] = pl.Series(rt_list)
    if with_ms1:
        mgf_data['ms1Intensity'] = pl.Series(ms1_intes)

    mgf_df = pl.DataFrame(mgf_data)
    mgf_df = mgf_df.unique(subset=[SOURCE_KEY, SCAN_KEY])

    return mgf_df



""" Generic functions for reading in any search results.
"""
import multiprocessing
import os

import pandas as pd
import polars as pl

from inspire.accession import process_accession_groups
from inspire.input.mascot import read_mascot_data
from inspire.input.maxquant import read_mq_data
from inspire.input.msfragger import read_ms_fragger_data
from inspire.input.peaks import read_peaks_data
from inspire.utils import add_fixed_modifications

def generic_read_df(config, save_dfs=True, for_calibration=False):
    """ Function to read in search results from any search engine.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the experiment.
    save_dfs : bool
        Flag indicating whether the formatted dataframes should be saved to disk.
    overwrite_reduce : bool
        Flag indicating whether to force reduction of Mascot dataframe to best hit
        (for CE calibration pipeline).

    Returns
    -------
    search_df : pd.DataFrame
        A DataFrame of search results.
    mods_df : pd.DataFrame
        A small DataFrame detailing the PTMs found.
    """
    if for_calibration:
        reduce_results = True
    else:
        reduce_results = config.rescore_method != 'percolatorSeparate'

    n_cores = min(config.n_cores, multiprocessing.cpu_count())

    if os.path.exists(f'{config.output_folder}/formated_df.csv') and config.reuse_input:
        search_df = pl.read_csv(f'{config.output_folder}/formated_df.csv')
        mods_df = pd.read_csv(f'{config.output_folder}/formated_mods.csv')
    else:
        if config.search_engine == 'mascot':
            search_df, mods_df = read_mascot_data(
                config.search_results,
                config.scan_title_format,
                config.source_files,
                reduce_results,
                config.source_filename,
                with_accession=config.use_accession_stratum,
            )
        elif config.search_engine == 'maxquant':
            search_df, mods_df = read_mq_data(config.search_results)
        elif config.search_engine == 'peaks':
            search_df, mods_df = read_peaks_data(config.search_results)
        elif config.search_engine == 'msfragger':
            search_df, mods_df = read_ms_fragger_data(
                config.search_results,
                config.fixed_modifications,
                n_cores,
                reduce_results,
            )
        else:
            raise ValueError(f'Unknown Search Engine: {config.search_engine}')
        
        if config.replace_il:
            search_df = search_df.with_columns(
                pl.col('peptide').str.replace_all('I', 'L')
            )
            search_df = search_df.unique(
                subset=['source', 'scan', 'peptide'], maintain_order=True
            )

        if config.use_accession_stratum:
            search_df = process_accession_groups(search_df, config)
        if (
            config.fixed_modifications is not None and
            config.search_engine != 'msfragger'
        ):
            search_df, mods_df = add_fixed_modifications(
                search_df,
                mods_df,
                config.fixed_modifications
            )

        if save_dfs and config.reuse_input:
            search_df.write_csv(f'{config.output_folder}/formated_df.csv')
            mods_df.to_csv(f'{config.output_folder}/formated_mods.csv', index=False)

    return search_df, mods_df



""" Script for shared utilities when plotting MS2 spectra.
"""
import numpy as np
import plotly.graph_objects as go
import plotly.io as pio
from plotly.subplots import make_subplots

from inspire.constants import (
    ION_OFFSET,
    KNOWN_PTM_WEIGHTS,
    LOSS_NAMES,
    NEUTRAL_LOSSES,
    PROTON,
)
from inspire.mz_match import get_ion_masses, compute_potential_mws


LOSS_NAMES = ['',  '*', '&#xb0;']


def experiment_match(exp_mzs, exp_intes, pred_mzs, plotting_names, mz_accuracy, mz_units):
    """ Function to match the mz values of the experimentally observed peaks to the
        Prosit predicted peaks.

    Parameters
    ----------
    exp_mzs : list of float
        The experimentally observed mz values.
    exp_intes : list of float
        The experimentally observed intensity values.
    pred_mzs : list of float
        The Prosit predicted mz values.

    Returns
    -------
    matched_mzs : list of float
        The matched mz values in the experimental spectrum.
    matched_intes : list of float
        The matched intensity values in the experimental spectrum.
    l2_norm : float
        The l2 norm of the matched intensities if more than 5 peaks are matched,
        otherwise the maximum intensity in the spectrum.
    matched_pred_mzs : list of float
        A list of the matched prosit predicted mz values.
    """
    matched_mzs, matched_intes = [], []
    matched_pred_mzs = []
    matched_names = []
    for match_name, pred_m in zip(plotting_names, pred_mzs):
        matched_ind = -1
        min_match = 100000
        for idx, exp_mz in enumerate(exp_mzs):
            if abs(exp_mz-pred_m) < min_match:
                min_match = abs(exp_mz-pred_m)
                matched_ind = idx

        if mz_units == 'ppm':
            mz_err = pred_m*mz_accuracy*(10**-6)
        else:
            mz_err = mz_accuracy

        if min_match < mz_err:
            matched_mzs.append(exp_mzs[matched_ind])
            matched_intes.append(exp_intes[matched_ind])
            matched_pred_mzs.append(pred_m)
            matched_names.append(match_name.split('+')[0])

    if len(matched_intes) > 2:
        l2_norm = np.linalg.norm(np.array(matched_intes), ord=2)
    else:
        l2_norm = np.max(exp_intes)

    matched_peaks = {
        'mzs': matched_mzs,
        'intensities': [x/l2_norm for x in matched_intes],
    }

    return matched_peaks, l2_norm, matched_pred_mzs, matched_names


def get_npp_ions(
        exp_mzs,
        exp_intes,
        matched_mzs,
        l2_norm,
        peptide,
        charge,
        ptm_id_weights,
        modifications,
        mz_accuracy,
        mz_units,
    ):
    """ Function to get the potential non-prosit-predicted ions.

    Parameters
    ----------
    exp_mzs : list of float
        A list of m/z values observed in the experimental spectrum.
    exp_intes : list of int
        A list of the intensities observed in the experimental spectrum.
    matched_mzs : list of float
        A list of the m/z values already matched to the Prosit predictions.
    l2_norm : float
        The L^2 norm of the Prosit matched ions in the experimental spectrum.
    peptide : string
        The peptide sequence of the precursor.
    charge : int
        The charge state of the peptide.
    ptm_id_weights : dict
        A dictionary mapping inSPIRE PTM IDs to their molecular weights.
    modifications : str
        The modifications of the peptide written as a string of inSPIRE IDs.
    mz_accuracy : float
        The accuracy of the mass spectrometer.
    mz_units : str
        Da or ppm

    Returns
    -------
    npp_ions : dict
        A dictionary of the m/z values, intensities, and names of non-prosit predicted ions.
    prec_ion : dict
        A dictionary of the m/z and intensity value of the precursor ion if observed in the MS2.
    """
    poss_mzs = []
    poss_intes = []
    poss_names = []
    prec_mz = []
    prec_inte = []
    all_mzs, names = generate_all_mzs(peptide, charge, ptm_id_weights, modifications)

    for idx, e_mz in enumerate(exp_mzs):
        if e_mz not in matched_mzs:
            if mz_units == 'ppm':
                mz_err = e_mz*mz_accuracy*(10**-6)
            else:
                mz_err = mz_accuracy

            for alt_idx, possible_mz in enumerate(all_mzs):
                if abs(e_mz-possible_mz) < mz_err:
                    if alt_idx == 0:
                        prec_mz.append(e_mz)
                        prec_inte.append(exp_intes[idx]/l2_norm)
                    poss_mzs.append(e_mz)
                    poss_intes.append(exp_intes[idx]/l2_norm)
                    poss_names.append(names[alt_idx])

    return {
        'mzs': poss_mzs,
        'intensities': poss_intes,
        'names': poss_names,
    }, {
        'mzs': prec_mz,
        'intensities': prec_inte,
    }

def generate_all_mzs(sequence, prec_charge, ptm_id_weights, modifications=None):
    """ Function to the get mz's for all the ions predicted by prosit.

    Parameters
    ----------
    sequence : str
        The peptide sequence for which we require molecular weights.
    modifications : str
        A string of the ptms for the sequence which will alter
        the potential mzs.

    Returns
    -------
    all_possible_ions : dict
        A dictionary of all the mzs of all possible b and y ions
        that could be produced.
    """
    sub_seq_mass, total_residue_mass = compute_potential_mws(
        sequence=sequence,
        modifications=modifications,
        reverse=False,
        ptm_id_weights=ptm_id_weights
    )

    rev_sub_seq_mass, _ = compute_potential_mws(
        sequence=sequence,
        modifications=modifications,
        reverse=True,
        ptm_id_weights=ptm_id_weights
    )

    # b- and y-ions for each of the charge options
    possible_ions = {}
    possible_ions['b'] = ION_OFFSET['b'] + sub_seq_mass
    possible_ions['y'] = ION_OFFSET['y'] + rev_sub_seq_mass
    possible_ions['a'] = ION_OFFSET['a'] + sub_seq_mass

    all_mzs = [(total_residue_mass+(prec_charge*PROTON))/prec_charge]
    names = ['precursor']

    for ion_type in 'bya':
        loss_list = [0.0] + NEUTRAL_LOSSES

        for frag_idx, weight in enumerate(possible_ions[ion_type]):
            for loss_idx, loss in enumerate(loss_list):
                for charge in range(1, prec_charge+1):
                    all_mzs.append(
                        ((weight-loss) + (PROTON*charge))/charge
                    )
                    charge_name = '+'*charge
                    names.append(
                        f'{ion_type}{frag_idx+1}{LOSS_NAMES[loss_idx]}{charge_name}'
                    )

    return all_mzs, names

def get_unmatched(exp_mzs, exp_intes, matched_mzs, npp_mzs, prec_mzs, l2_norm):
    """ Function select and normalise the unmatched peaks from the experimental spectrum.

    Parameters
    ----------
    exp_mzs : list of float
        List of the mz values of the experimental spectrum.
    exp_intes : list of float
        List of the intensity values of the experimental spectrum.
    matched_mzs : list of float
        List of the mz values matched to the Prosit spectrum.
    l2_norm : float
        Either the l2 norm of the matched spectrum or the maximum intensity in the
        experimental spectrum if fewer than 5 peaks are matched.

    Returns
    -------
    unmatched_mzs : list of float
        List of the mz values of the experimental spectrum not matched
        to the Prosit predicted spectrum.
    exp_intes : list of float
        List of the intensity values of the experimental spectrum not
        matched to the Prosit predicted spectrum.
    """
    unmatched_mzs, unmatched_intes = [], []
    for idx, e_mz in enumerate(exp_mzs):
        if e_mz not in matched_mzs and e_mz not in npp_mzs and e_mz not in prec_mzs:
            unmatched_mzs.append(e_mz)
            unmatched_intes.append(exp_intes[idx]/l2_norm)

    return {
        'mzs': unmatched_mzs,
        'intensities': unmatched_intes,
    }


def get_plot_details(
        index, pred_mzs, matched_p_mz, pred_intes, plotting_names,
        peptide, matched_names, non_prosit_pred_peaks
    ):
    """ Function to get the traces, colours and annotations shown in the MS2 spectral
        plot.
    """
    pep_len = len(peptide)
    annotations = []
    colours = []
    for idx, entry in enumerate(pred_mzs):
        if entry in matched_p_mz:
            colour = '#0095A8'
        else:
            colour = '#FF7043'
        colours.append(colour)

        if pred_intes[idx] < -0.05:
            annotations.append(
                {
                    'x': entry,
                    'ax': entry,
                    'y': pred_intes[idx]-0.05,
                    'ay': pred_intes[idx]-0.05,
                    'xref': f'x{index+1}',
                    'yref': f'y{index+1}',
                    'text': plotting_names[idx],
                    'font_size': 8,
                    'font_color': colour,
                    'showarrow': False,
                }
            )

    if pep_len < 20:
        annot_height = 1.5
    else:
        annot_height = 0.7

    extra_traces = []
    annotations.append({
        'showarrow': False,
            'text': 'Experimental Spectrum',
            'x': 1300,
            'ax': 1300,
            'y': annot_height,
            'ay': annot_height,
            'font_size': 12,
            'font_family': "Arial, monospace",
            'xref': f'x{index+1}',
            'yref': f'y{index+1}',
            'align': 'left',
    })
    annotations.append({
        'showarrow': False,
            'text': 'Prosit Predicted Spectrum',
            'x': 1300,
            'ax': 1300,
            'y': -1.1,
            'ay': -1.1,
            'font_size': 12,
            'font_family': "Arial, monospace",
            'xref': f'x{index+1}',
            'yref': f'y{index+1}',
            'align': 'left',
    })
    for idx, residue in enumerate(peptide):
        annotations.append({
            'showarrow': False,
            'text': (residue),
            'x': 50 + (idx * 50),
            'ax': 50 + (idx * 50),
            'y': 1.25,
            'ay': 1.25,
            'font_size': 12,
            'font_family': "Arial, monospace",
            'xref': f'x{index+1}',
            'yref': f'y{index+1}',
            'align': 'left',
        })
        if idx > 0:
            min1_idx = idx-1
            rev_idx = len(peptide) - idx
            y_matched = False
            b_matched = False
            npp_names = non_prosit_pred_peaks['names']
            if f'b{idx}' in matched_names:
                annotations.append({
                    'showarrow': False,
                    'text': f'b{idx}',
                    'x': 40 + (min1_idx * 50),
                    'ax': 40 + (min1_idx * 50),
                    'y': 1.02,
                    'ay': 1.02,
                    'font_size': 10,
                    'font_family': "Arial, monospace",
                    'xref': f'x{index+1}',
                    'yref': f'y{index+1}',
                    'align': 'left',
                })
                b_matched = True
            if f'y{rev_idx}' in matched_names:
                annotations.append({
                    'showarrow': False,
                    'text': f'y{rev_idx}',
                    'x': 63 + (idx * 50),
                    'ax': 63 + (idx * 50),
                    'y': 1.49,
                    'ay': 1.49,
                    'font_size': 10,
                    'font_family': "Arial, monospace",
                    'xref': f'x{index+1}',
                    'yref': f'y{index+1}',
                    'align': 'left',
                })
                y_matched = True
            for name in npp_names:
                if name == 'precursor':
                    continue
                frag_name = name.split('&')[0].split('*')[0].split('+')[0]
                frag_type = frag_name[0]
                frag_idx = int(frag_name[1:])
                if rev_idx == frag_idx:
                    if frag_type == 'y':
                        if not y_matched:
                            annotations.append({
                                'showarrow': False,
                                'text': name.split('+')[0],
                                'x': 63 + (idx * 50),
                                'ax': 63 + (idx * 50),
                                'y': 1.49,
                                'ay': 1.49,
                                'font_size': 10,
                                'font_family': "Arial, monospace",
                                'xref': f'x{index+1}',
                                'yref': f'y{index+1}',
                                'align': 'left',
                            })
                        y_matched = True
                if idx == frag_idx:
                    if frag_type != 'y':
                        if not b_matched:
                            annotations.append({
                                'showarrow': False,
                                'text': name.split('+')[0],
                                'x': 40 + (min1_idx * 50),
                                'ax': 40 + (min1_idx * 50),
                                'y': 1.02,
                                'ay': 1.02,
                                'font_size': 10,
                                'font_family': "Arial, monospace",
                                'xref': f'x{index+1}',
                                'yref': f'y{index+1}',
                                'align': 'left',
                            })
                        b_matched = True

        if idx > 0:
            if b_matched:
                extra_traces.append(
                    go.Scatter(
                        x=[50 + ((idx-1) * 50), 50 + ((idx-0.5) * 50)],
                        y=[1.1, 1.25],
                        mode='lines',
                        line_color='black',
                        line_width=0.5,
                    )
                )
            if y_matched:
                extra_traces.append(
                    go.Scatter(
                        x=[50 + ((idx-0.5) * 50), 50 + (idx * 50)],
                        y=[1.25, 1.4],
                        mode='lines',
                        line_color='black',
                        line_width=0.5,
                    )
                )
            if b_matched:
                extra_traces.append(
                    go.Scatter(
                        x=[40 + ((idx-1) * 50), 50 + ((idx-1) * 50)],
                        y=[1.1, 1.1],
                        mode='lines',
                        line_color='black',
                        line_width=0.5,
                    )
                )
            if y_matched:
                extra_traces.append(
                    go.Scatter(
                        x=[50 + ((idx) * 50), 60 + (idx * 50)],
                        y=[1.4, 1.4],
                        mode='lines',
                        line_color='black',
                        line_width=0.5,
                    )
                )

    return annotations, colours, extra_traces

def create_traces(
        pred_mzs, pred_intes, unmatched_peaks, non_prosit_pred_peaks,
        prec_peak, matched_peaks, colours,
    ):
    """ Function to create the required bar traces for MS2 spectral plotting.
    """
    return [
        go.Bar(
            x=pred_mzs,
            base='relative',
            alignmentgroup='predicted',
            y=pred_intes,
            name='Predicted',
            marker_color=colours,
            textfont={
                'size': 40,
                'color': 'black'
            },
            marker_line_width=0,
            width=4,
            textposition='outside',
        ),
        go.Bar(
            x=unmatched_peaks['mzs'],
            y=unmatched_peaks['intensities'],
            base='relative',
            alignmentgroup='experimental',
            name='Experimental Not Matched',
            marker_color='#E8E8E8',
            marker_line_width=0,
            width=4,
        ),
        go.Bar(
            x=non_prosit_pred_peaks['mzs'],
            y=non_prosit_pred_peaks['intensities'],
            base='relative',
            alignmentgroup='experimental',
            name='Possible Ion Not Predicted by Prosit',
            marker_color='darkseagreen',
            marker_line_width=0,
            width=4,
        ),
        go.Bar(
            x=prec_peak['mzs'],
            y=prec_peak['intensities'],
            base='relative',
            alignmentgroup='experimental',
            name='Precursor',
            marker_color='plum',
            marker_line_width=0,
            width=4,
        ),
        go.Bar(
            x=matched_peaks['mzs'],
            y=matched_peaks['intensities'],
            base='relative',
            alignmentgroup='experimental',
            name='Experimental Matched',
            marker_color='black',
            marker_line_width=0,
            width=4,
        ),
    ]



def convert_names_and_mzs(mod_seq, pred_names):
    """ Function to generate plotting ion names and mzs.

    Parameters
    ----------
    mod_seq : str
        The Prosit modified_sequence.
    pred_names : list of str
        A list of the names of the ions as read in from msp format.

    Returns
    -------
    pred_mzs : list of float
        A list of the theoretical mz values for the prosit predicted ions.
    plotting_names : list of str
        A list of the annotations for the prosit predicted ions.
    """
    if '(ox)' not in mod_seq and 'C' not in mod_seq:
        mods = None
        un_mod_seq = mod_seq.replace('(ox)', '')
    else:
        mods = '0.'
        un_mod_seq = ''
        while mod_seq:
            if len(mod_seq) > 1:
                if mod_seq[1] == '(':
                    un_mod_seq += 'M'
                    mods += '1'
                    mod_seq = mod_seq[5:]
                    continue
            un_mod_seq += mod_seq[0]
            if mod_seq[0] != 'C':
                mods += '0'
            else:
                mods += '2'
            mod_seq = mod_seq[1:]
        mods += '.0'
    masses, _ = get_ion_masses(
        un_mod_seq,
        {
            0: 0.0,
            1: KNOWN_PTM_WEIGHTS['Oxidation (M)'],
            2: KNOWN_PTM_WEIGHTS['Carbamidomethylation'],
        },
        modifications=mods
    )
    pred_mzs = []
    plotting_names = []
    for pred_ion in pred_names:
        ion_data = pred_ion.split('^')
        if '^' in pred_ion:
            charge = int(ion_data[1])
        else:
            charge = 1
        plotting_names.append(ion_data[0] + ('+' * charge))
        code = ion_data[0][0]
        idx = int(ion_data[0][1:])
        pred_mzs.append(
            (masses[code][idx-1] + (PROTON*charge))/charge
        )
    return pred_mzs, plotting_names



def add_legend(output_folder, experiment_title, plot_type):
    """ Function for adding a legend to the start of the spectrolPlots.pdf file.

    Parameters
    ----------
    output_folder : str
        The folder where all inSPIRE outputs are written.
    experiment_title : str
        The title of the experiment.
    plot_type : str
        standard or isobar
    """
    legend_traces = [
        go.Scatter(
            x=[1.0],
            y=[7.5],
            marker={'size':22, 'color':'black'},
            text='    Experimental peak matched to a Prosit predicted peak.',
            textposition='middle right',
            textfont_size=22,
            textfont_family='Helvetica',
            mode='markers+text',
        ),
        go.Scatter(
            x=[1.0],
            y=[5.0],
            marker={'size':22, 'color':'darkseagreen'},
            text='    Possible ion unknown to Prosit.',
            textposition='middle right',
            textfont_size=22,
            textfont_family='Helvetica',
            mode='markers+text',
        ),
        go.Scatter(
            x=[1.0],
            y=[2.5],
            marker={'size':22, 'color':'plum'},
            text='    Precursor matched peak.',
            textposition='middle right',
            textfont_size=22,
            textfont_family='Helvetica',
            mode='markers+text',
        ),
        go.Scatter(
            x=[1.0],
            y=[0.0],
            marker={'size':22, 'color':'#E8E8E8'},
            text='    Experimental peak not matched to any potential ion.',
            textposition='middle right',
            textfont_size=22,
            textfont_family='Helvetica',
            mode='markers+text',
        ),
        go.Scatter(
            x=[1.0],
            y=[7.5],
            marker={'size':22, 'color':'#0095A8'},
            text='    Prosit predicted peak matched to experimental spectrum.',
            textposition='middle right',
            textfont_size=22,
            textfont_family='Helvetica',
            mode='markers+text',
        ),
        go.Scatter(
            x=[1.0],
            y=[5.0],
            marker={'size':22, 'color':'#FF7043'},
            text='    Prosit predicted peak not matched to experimental spectrum.',
            textposition='middle right',
            textfont_size=22,
            textfont_family='Helvetica',
            mode='markers+text',
        ),
    ]

    colour_key_fig = make_subplots(
        rows=5,
        cols=1,
        subplot_titles=[
            '',
            'Experimental Spectrum Colour Code:',
            'Prosit Spectrum Colour Code:',
            'Additional Notes:'
        ],
    )
    for trace in legend_traces[:4]:
        colour_key_fig.add_trace(trace, row=2, col=1)
    for trace in legend_traces[4:]:
        colour_key_fig.add_trace(trace, row=3, col=1)

    colour_key_fig.add_trace(
        go.Scatter(
            x=[1.0],
            y=[7.5],
            text='&#xb0; indicates an ion with loss of H<sub>2</sub>O.',
            textposition='middle right',
            textfont_size=22,
            textfont_family='Helvetica',
            mode='text',
        ),
        row=4,
        col=1,
    )
    colour_key_fig.add_trace(
        go.Scatter(
            x=[1.0],
            y=[5.0],
            text='* indicates an ion with loss of NH<sub>3</sub>.',
            textposition='middle right',
            textfont_size=22,
            textfont_family='Helvetica',
            mode='text',
        ),
        row=4,
        col=1,
    )

    colour_key_fig.update_layout(
        width=2100,
        height=2500,
        paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
        showlegend=False,
    )


    colour_key_fig.update_xaxes(
        showticklabels=True,
        range=[0, 4],
        dtick=300,
        linecolor='black',
        linewidth=0.5,
        showgrid=False,
        ticks="outside",
        zeroline=False, # thick line at x=0
        visible=False
    )

    colour_key_fig.update_yaxes(
        showline=True,
        linewidth=0.5,
        linecolor='black',
        range=[-2, 10],
        tickvals = [-1.2+(i*0.4) for i in range(8)],
        ticktext = [round(abs(-1.2+(i*0.4)), 1) for i in range(8)],
        showgrid=False,
        ticks="outside",
        zeroline=False, # thick line at x=0
        visible=False
    )

    colour_key_fig.update_annotations(
        {
            'font_size': 26,
            'font_family': "Helvetica",
        }
    )

    if plot_type == 'standard':
        title_text = f'<b>inSPIRE Spectral Plotting for {experiment_title}</b>'
    else:
        title_text = f'<b>inSPIRE Spectral Plotting for {experiment_title}</b>'

    colour_key_fig.update_layout(
        title_text=title_text,
        title_x=0.5,
        title_font_size=30,
        title_font_family='Helvetica',
    )

    if plot_type == 'standard':
        out_path = f'{output_folder}/spectralPlots_legend.pdf'
    else:
        out_path = f'{output_folder}/spectralPlots_legend.pdf'

    pio.write_image(
        colour_key_fig,
        out_path,
        engine='kaleido',
    )


def update_fig_layout(fig, n_plot_rows):
    """ Function to update the figure layout.
    """
    fig.update_layout(
        width=2100,
        height=n_plot_rows*500,
        paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
        showlegend=False,
    )


    fig.update_xaxes(
        showticklabels=True,
        range=[0, 1500],
        dtick=300,
        linecolor='black',
        linewidth=0.5,
        showgrid=False,
        ticks="outside",
    )

    fig.update_yaxes(
        showline=True,
        linewidth=0.5,
        linecolor='black',
        range=[-1.2, 1.6],
        tickvals = [-1.2+(i*0.4) for i in range(8)],
        ticktext = [round(abs(-1.2+(i*0.4)), 1) for i in range(8)],
        showgrid=False,
        ticks="outside",
    )

    return fig



""" Functions for reading in or output of .ssl files.
"""
import re

import pandas as pd

from inspire.utils import remove_source_suffixes

def ssl_file_to_inspire_format(ssl_loc):
    """ Function to read in data from .ssl format

    Parameters
    ----------
    ssl_loc : str
        Location of the ssl file.

    Returns
    -------
    pd.DataFrame : pd.DataFrame
        DataFrame from ssl data read in.
    """
    id_df = pd.read_csv(ssl_loc, sep='\t')
    id_df = id_df.rename(
        columns={
            'file': 'source',
            'sequence': 'peptide',
            'modifications': 'modifiedSequence',
        },
    )
    id_df['source'] = id_df['source'].apply(remove_source_suffixes)
    id_df['peptide'] = id_df['peptide'].apply(lambda x : re.sub( r"[^a-zA-Z]+", "", x))

    return id_df



""" Functions for running automated feature selection.
"""
from itertools import combinations

import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import auc, precision_recall_curve
from sklearn.model_selection import train_test_split
import yaml

from inspire.constants import (
    CHARGE_KEY,
    DELTA_SCORE_KEY,
    ENDC_TEXT,
    ENGINE_SCORE_KEY,
    FRAG_MZ_ERR_MED_KEY,
    FRAG_MZ_ERR_VAR_KEY,
    LABEL_KEY,
    LOSS_IONS_KEY,
    MASS_DIFF_KEY,
    MINIMAL_FEATURE_SET,
    NOT_ASSIGNED_KEY,
    OKCYAN_TEXT,
    PEARSON_KEY,
    PEPTIDE_KEY,
    PRECURSOR_INTE_KEY,
    PREFIX_KEYS,
    SPEARMAN_KEY,
    SUFFIX_KEYS,
    SEQ_LEN_KEY,
    SOURCE_INDEX_KEY,
    SOURCE_KEY,
    SPECTRAL_ANGLE_KEY,
)
from inspire.spectral_features import DELTA_FEATURES

BASE_FEATURES = [
    SPECTRAL_ANGLE_KEY,
    DELTA_SCORE_KEY,
    ENGINE_SCORE_KEY,
]

FIRST_ADDITIONS = [
    'deltaRT',
    SEQ_LEN_KEY,
]

SPEARMAN_METRICS = [
    SPEARMAN_KEY,
    'spearmanMinorIons',
    'spearmanMajorIons',
    'medianAbsoluteError',
    PEARSON_KEY,
]

COVERAGE_FEATURES = [
    'matchedCoverage',
    'minMatchedCoverage',
    'maxMatchedCoverage',
    'prositDeltaQuartile1',
    'prositDeltaQuartile3',
    'prositDeltaMedian',
    'nDeltasAboveThreshold',
    'nDeltasAboveZero',
    'minPrositDelta',
    'maxPrositDelta',
]

AA_FEATS = [
    'possibleKrFragsDivTotal',
    'fracMatchedKR',
    'fracKR',
    'fracC',
    'missedCleavages',
    'nVarMods',
    'fracUnique',
    'nRepeatedResidues',
    'seqLenMeanDiff',
]

INTENSITY_FEATURES = [
    'nMajorNotMatchableDivFrags',
    'nMinorNotMatchableDivFrags',
    PRECURSOR_INTE_KEY,
    'nMajorMatchedDivFrags',
    'nMinorMatchedDivFrags',
]

MZ_DISTRIBUTION = [
    CHARGE_KEY,
    'avgResidueMass',
    MASS_DIFF_KEY,
    FRAG_MZ_ERR_MED_KEY,
    FRAG_MZ_ERR_VAR_KEY,
    'spectrumDensity',
    LOSS_IONS_KEY,
    'fromChimera',
]



FOUND_NOT_PRED_FEATS = [
    NOT_ASSIGNED_KEY,
    'nMajorPredNotFoundDivFrags',
]


YB_DIFF_FEATS = [
    'maxTypeSpectralAngle',
    'minTypeSpectralAngle',
    'yIsDominantIonSeries',
    'bIsDominantIonSeries',
]


SPECTRAL_FEATURE_GROUPS = [
    SPEARMAN_METRICS,
    COVERAGE_FEATURES,
    AA_FEATS,
    INTENSITY_FEATURES,
    MZ_DISTRIBUTION,
    FOUND_NOT_PRED_FEATS,
    YB_DIFF_FEATS,
]

SPECTRAL_GROUP_NAMES = [
    'Spearman R Features',
    'Coverage Features',
    'Amino Acid Features',
    'Matched/Unmatched Intensity Features',
    'MZ Distribution Features',
    'Found but not Predicted Features',
    'Y/B Ion Difference Features',
]





DEFAULT_FEATURE_SET = [
    ENGINE_SCORE_KEY,
    'deltaScore',
    'sequenceLength',
    'seqLenMeanDiff',
    'charge',
    'avgResidueMass',
    'nVarMods',
    'spectralAngle',
    'deltaRT',
    SPEARMAN_KEY,
    PEARSON_KEY,
    'spearmanMajorIons',
    'medianAbsoluteError',
    'matchedCoverage',
    'maxMatchedCoverage',
    'minMatchedCoverage',
    'minPrositDelta',
    'prositDeltaQuartile1',
    'prositDeltaMedian',
    'prositDeltaQuartile3',
    'maxPrositDelta',
    'nDeltasAboveZero',
    'nDeltasAboveThreshold',
    'fracUnique',
    'nRepeatedResidues',
    'medianFragmentMzError',
    'fragmentMzErrorVariance',
    'nMajorMatchedDivFrags',
    'nMinorMatchedDivFrags',
    'nMajorNotMatchableDivFrags',
    'nMinorNotMatchableDivFrags',
    'maxTypeSpectralAngle',
    'yIsDominantIonSeries',
    LOSS_IONS_KEY,
    'predNotFoundCoverage',
    'fracC',
    'fracKR',
    'fracMatchedKR',
    'spectrumDensity',
    'fromChimera',
    'missedCleavages',
]

OUTSTANDING_CONFOUNDING_VARIABLES = [
    SOURCE_INDEX_KEY,
]

THRESHOLD_VALUE = 20


def get_best_feature_set(fe_df, feature_set):
    """ Function to get the best combination of features from a feature set
        for each number of features.

    Parameters
    ----------
    fe_df : pd.DataFrame
        The DataFrame containg all features.
    feature_set : list of str
        A list of the features we wish to select between.

    Returns
    -------
    features_sets_to_test : list of list of str
        A list of the best feature combinations for each number of features.
    """
    fe_train, fe_test = train_test_split(fe_df, test_size=0.33, random_state=42)
    fe_train = fe_train[fe_train['testMetric'] != -1]
    fe_test['trueLabel'] = fe_test['testMetric'].apply(lambda x : 1 if x == -1 else x)
    features_sets_to_test = []
    for i in range(1, len(feature_set)+1):
        max_r2 = -1000
        best_features = []
        for feature_selection in combinations(feature_set, i):
            feature_selection = list(feature_selection)
            reg = LogisticRegression(max_iter=1000, random_state=42).fit(
                fe_train[feature_selection + BASE_FEATURES],
                fe_train['testMetric'],
            )

            preds = reg.predict_proba(
                fe_test[feature_selection + BASE_FEATURES]
            )[:, 1]

            precision, recall, _ = precision_recall_curve(fe_test['trueLabel'], preds)
            feat_set_r2 = auc(recall, precision)

            if feat_set_r2 > max_r2:
                max_r2 = feat_set_r2
                best_features = feature_selection

        features_sets_to_test.append(best_features)
    return features_sets_to_test


def remove_excluded_features(feature_list, all_features_df, exclude_features):
    """ Function to get the maximum possible number of features for Percolator q-value
        calculation when dealing with small datasets.

    Parameters
    ----------
    feature_list : list of str
        A list of perspective features.
    all_features_df : pd.DataFrame
        A DataFrame of perspective Percolator inputs.
    exclude_features : list of str
        A list of features to be ignored.

    Returns
    -------
    feature_list : list of str
        A list of perspective features without excluded features.
    """
    feature_list = [
        x for x in feature_list if (
            x not in exclude_features and all_features_df[x].nunique() > 1
        )
    ]
    return feature_list

def create_test_metric(df_row, sa_cut_off, es_cut_off):
    """ Function to create the metric on which logistic models will be trained to
        quickly test feature set performance.

    Parameters
    ----------
    df_row : pd.Series
        The row of a DataFrame in which we wish to create a testMetric.

    Returns
    -------
    df_row : pd.Series
        The updated DataFrame row with testMetric added.
    """
    if (
        df_row[LABEL_KEY] > 0 and
        df_row[SPECTRAL_ANGLE_KEY] > sa_cut_off and
        df_row[ENGINE_SCORE_KEY] > es_cut_off
    ):
        return 1
    if df_row[LABEL_KEY] > 0:
        return -1
    return 0


def convert_to_irt(all_features_df, config):
    """ Function to convert deltaRT in seconds to iRT value.
    """
    all_features_df[SOURCE_KEY] = all_features_df['specID'].apply(
        lambda x : '_'.join(x.split('_')[:-2])
    )
    sources = all_features_df[SOURCE_KEY].unique().tolist()
    irt_coeffs = {}
    for source in sources:
        try:
            irt_df = pd.read_csv(f'{config.output_folder}/rt_fit_{source}.csv')
            irt_coeffs[source] = irt_df['coefficents'].mean()
            if irt_coeffs[source] <= 0:
                irt_coeffs[source] = 1.0
        except FileNotFoundError:
            print(f'No file found for {source}')
            irt_coeffs[source] = 1.0

    all_features_df['deltaRT'] = all_features_df[['deltaRT', 'source']].apply(
        lambda df_row : abs(
            df_row['deltaRT']/irt_coeffs.get(df_row['source'], 1.0)
        ),
        axis=1,
    )
    return all_features_df


def select_features(config):
    """ Function to select the features used in the final percolator model.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the experiment.
    """
    all_features_df = pd.read_csv(
        f'{config.output_folder}/input_all_features.tab',
        sep='\t'
    )

    if config.use_irt_diff:
        all_features_df = convert_to_irt(all_features_df, config)

    if config.minimal_features:
        feature_set = MINIMAL_FEATURE_SET
    else:
        feature_set = DEFAULT_FEATURE_SET
        if config.delta_method == 'ignore':
            feature_set = [x for x in feature_set if x not in DELTA_FEATURES]

    if config.use_binding_affinity == 'asFeature':
        feature_set += ['bindingAffinity']

    if config.use_accession_stratum:
        feature_set += [
            col for col in all_features_df.columns # pylint: disable=no-member
            if col.startswith('accession_')
        ]

    if config.exclude_features is not None and config.exclude_features:
        exclude_features = (
            config.exclude_features +
            SUFFIX_KEYS[config.rescore_method] +
            PREFIX_KEYS[config.rescore_method]
        )
    elif config.include_features is not None:
        feature_set = list(set(feature_set + config.include_features))
        exclude_features = [
            col for col in all_features_df.columns # pylint: disable=no-member
            if col not in config.include_features
        ]
        if config.use_accession_stratum:
            exclude_features = [x for x in exclude_features if not x.startswith('accession')]
    else:
        exclude_features = (
            SUFFIX_KEYS[config.rescore_method] + PREFIX_KEYS[config.rescore_method]
        )

    feature_set = remove_excluded_features(
        feature_set, all_features_df, exclude_features
    )

    write_final_feature_set(all_features_df, feature_set, config)

def write_final_feature_set(all_features_df, feature_set, config):
    """ Function to write the final selected features for Percolator input.

    Parameters
    ----------
    all_features_df : pd.DataFrame
        DataFrame containing all possible percolator input features.
    feature_set : list of str
        A list of the feature names to be used.
    config : inspire.config.Config
        The Config object
    """
    with open(f'{config.output_folder}/selectedFeatures.yaml', 'w', encoding='UTF-8') as file:
        yaml.dump(feature_set, file)

    prefix_keys = PREFIX_KEYS[config.rescore_method]

    all_features_df[PEPTIDE_KEY] = '-.' +all_features_df[PEPTIDE_KEY].astype(str) + '.-'
    all_features_df = all_features_df[
        prefix_keys + feature_set + SUFFIX_KEYS[config.rescore_method]
    ]

    all_features_df.to_csv(
        f'{config.output_folder}/final_input.tab',
        sep='\t',
        index=False
    )
    print(
        OKCYAN_TEXT +
        '\tFinal Feature Set Written.' +
        ENDC_TEXT
    )



""" Functions to perform differential expression analysis between infected and
    control measurements.
"""
from math import log2, ceil

import numpy as np
import pandas as pd
import plotly.express as px
from scipy.stats import ttest_ind
from sklearn.cluster import KMeans
from statsmodels.sandbox.stats.multicomp import multipletests

from inspire.constants import ACCESSION_KEY, PEPTIDE_KEY
from inspire.utils import is_control
from inspire.quant.utils import create_quant_logoplots

FC_CUT = 0.5
P_VAL_CUT = 0.05
LOG_P_VAL_CUT = -log2(0.05)


def de_analysis(config):
    """ Function to perform differential expression analysis between infected and control
        measurements.

    Parameters
    ----------
    config : inspire.config.Config
        Config object for the whole experiment.
    """
    if config.control_flags is None:
        return

    quant_df = pd.read_csv(
        f'{config.output_folder}/quant/normalised_quantification.csv'
    )

    quant_df, ctrl_samp_names, inf_samp_names = combine_samples(quant_df, config)

    quant_df['nValidControl'] = quant_df[ctrl_samp_names].count(axis=1)
    quant_df['nValidInf'] = quant_df[inf_samp_names].count(axis=1)

    quant_df = calculate_fold_change_and_p_val(quant_df, ctrl_samp_names, inf_samp_names)


    # Write outputs.
    quant_df['absFoldChange'] = quant_df['foldChange'].abs()
    quant_df = quant_df.sort_values(by='absFoldChange', ascending=False)


    quant_df[[
        PEPTIDE_KEY, ACCESSION_KEY,
    ] + ctrl_samp_names + inf_samp_names + [
        'foldChange',
        'pValue',
        'adjusted_pValue',
        'diffExprResult',
    ]].to_csv(f'{config.output_folder}/quant/de_peptide_results.csv', index=False)

    quant_df = quant_df[quant_df['foldChange'].notna()]
    kmeans = KMeans(n_clusters=7, random_state=42)
    quant_df['cluster'] = kmeans.fit_predict(quant_df[['foldChange']])

    
    try:
        create_quant_logoplots(quant_df, config.output_folder, kmeans.cluster_centers_)
    except Exception as e:
        print(f'Quantitative logo plotting failed with exception {e}')

    quant_df = quant_df[
        (quant_df['nValidControl'] >= 3) &
        (quant_df['nValidInf'] >= 3)
    ]
    if quant_df.shape[0]:
        create_volcano_plot(quant_df, config)

def combine_samples(quant_df, config):
    """ Function to combine technical replicate values into an averaged sample value.
    """
    meta_df = pd.read_csv(f'{config.output_folder}/quant/metadata.csv')
    file_name_mapping = dict(zip(
        meta_df['source'].tolist(), meta_df['renamed'].tolist()
    ))

    ctrl_samples = 0
    inf_samples = 0
    sample_meta_list = []
    sample_names = []
    ctrl_samp_names = []
    inf_samp_names = []
    if config.technical_replicates is None:
        config.technical_replicates = [[raw_file] for raw_file in meta_df['source'].tolist()]
    for sample_group in config.technical_replicates:
        if is_control(sample_group[0], config.control_flags):
            ctrl_samples += 1
            name = f'controlSample{ctrl_samples}'
            ctrl_samp_names.append(name)
        else:
            inf_samples += 1
            name = f'infectionSample{inf_samples}'
            inf_samp_names.append(name)

        quant_df[name] = quant_df[
            [f'{file_name_mapping[raw_file]}_norm' for raw_file in sample_group]
        ].mean(axis=1)

        for raw_file in sample_group:
            sample_meta_list.append(
                {'source': raw_file, 'sample': name}
            )
        sample_names.append(name)

    # Add the sample to the metadata
    if 'sample' in meta_df.columns:
        meta_df = meta_df.drop('sample', axis=1)

    sample_meta_df = pd.DataFrame(sample_meta_list)
    meta_df = pd.merge(meta_df, sample_meta_df, on='source', how='inner')
    meta_df.to_csv(f'{config.output_folder}/quant/metadata.csv', index=False)

    quant_df = quant_df[[PEPTIDE_KEY, ACCESSION_KEY] + sample_names]

    return quant_df, ctrl_samp_names, inf_samp_names


def calculate_fold_change_and_p_val(quant_df, ctrl_samp_names, inf_samp_names):
    """ Function to calculate the fold change and p-values for difference between infected
        and control samples.

    Parameters
    ----------
    quant_df : pd.DataFrame
        DataFrame of normalised quantified peptides.
    ctrl_samp_names : list of str
        A list of the sample names for control measurements.
    inf_samp_names : list of str
        A list of the sample names for infected measurements.

    Returns
    -------
    quant_df : pd.DataFrame
        The input DataFrame with differential expression analysis performed.
    """
    # Get geometric mean values:
    quant_df['meanControl'] = quant_df[ctrl_samp_names].mean(axis=1)
    quant_df['meanInfected'] = quant_df[inf_samp_names].mean(axis=1)

    # Calculate fold change and p-value
    quant_df['foldChange'] = quant_df['meanInfected'] - quant_df['meanControl']
    quant_df['pValue'] = quant_df.apply(
        lambda df_row : ttest_ind(
            [df_row[x] for x in ctrl_samp_names if not np.isnan(df_row[x])],
            [df_row[x] for x in inf_samp_names if not np.isnan(df_row[x])],
            equal_var=True,
        )[1],
        axis=1,
    )

    quant_df['-lg2(pValue)'] = quant_df['pValue'].apply(lambda x : -log2(x))

    if not quant_df.shape[0]:
        return quant_df

    quant_df['adjusted_pValue'] = multipletests(quant_df['pValue'], method='fdr_bh')[1]

    # Define results for volcano plot.
    quant_df['diffExprResult'] = quant_df[['pValue', 'foldChange']].apply(
        lambda df_row : 'Not Significant' if df_row['pValue'] >= P_VAL_CUT or abs(
            df_row['foldChange']
        ) < FC_CUT else (
            'Upregulated' if df_row['foldChange'] > 0 else 'Downregulated'
        ),
        axis=1
    )

    return quant_df


def create_volcano_plot(quant_df, config):
    """ Function to create a volcano plot based on the peptide intensity changes.

    Parameters
    ----------
    quant_df : pd.DataFrame
        A DataFrame of quantified peptides with DE analysis performed.
    config : inspire.config.Config
        Config object for the whole experiment.
    """
    # Scatter plot fold change and -lg of p value
    fig = px.scatter(
        quant_df,
        x='foldChange',
        y='-lg2(pValue)',
        color='diffExprResult',
        color_discrete_map={
            'Downregulated': 'royalblue',
            'Upregulated': 'tomato',
            'Not Significant': 'grey',
        },
        hover_name='peptide',
        hover_data='peptide',
    )
    fig.update_traces(marker={'size': 4})

    # Add lines for significance cut-offs
    fig.add_hline(y=LOG_P_VAL_CUT, line_color='black', line_dash='dash', line_width=0.5)
    fig.add_vline(x=FC_CUT, line_color='black', line_dash='dash', line_width=0.5)
    fig.add_vline(x=-FC_CUT, line_color='black', line_dash='dash', line_width=0.5)

    # Calculate the range over which fold changes should be plotted
    fc_max = quant_df['foldChange'].max()
    fc_min = quant_df['foldChange'].min()
    fc_range = ceil(max(fc_max, abs(fc_min)))
    if fc_range % 2:
        fc_range += 1

    # Clean axes and layout
    fig.update_xaxes(
        showticklabels=True,
        range=[-fc_range, fc_range],
        title_text='Log Fold Change',
        linecolor='black',
        linewidth=0.5,
        showgrid=False,
        ticks="outside",
    )
    fig.update_yaxes(
        showline=True,
        showticklabels=True,
        linewidth=0,
        linecolor='black',
        title_text='-log2(pValue)',
        range=[0,ceil(quant_df['-lg2(pValue)'].max())],
        showgrid=False,
        ticks="outside",
    )
    fig.update_layout(
        width=500,
        height=300,
        paper_bgcolor='rgba(256,256,256,256)',
        plot_bgcolor='rgba(256,256,256,256)',
        showlegend=False,
        font_color='black',
        font_family='Helvetica',
    )

    # Save image as svg
    fig.write_image(
        f'{config.output_folder}/img/peptide_volcano.svg', engine='kaleido'
    )



""" Functions for plotting spectra.
"""
import os

from PyPDF2 import PdfMerger
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.io as pio

from inspire.constants import (
    CHARGE_KEY,
    KNOWN_PTM_WEIGHTS,
    PEPTIDE_KEY,
    SCAN_KEY,
    SOURCE_KEY,
    SPECTRAL_ANGLE_KEY,
)

from inspire.input.msp import msp_to_df
from inspire.predict_spectra import predict_spectra
from inspire.spectral_features import calculate_spectral_features
from inspire.utils import convert_mod_seq_to_ptm_seq, fetch_scan_data
from inspire.plot_spectra.plot_spec_utils import (
    convert_names_and_mzs, create_traces, get_plot_details, get_unmatched,
    experiment_match, get_npp_ions, add_legend, update_fig_layout,
)

PLOTS_PER_PAGE = 15
PLOTS_PER_LINE = 3


def pair_plot(df_row, mz_accuracy, mz_units):
    """ Function to generate the traces and annotations needed for the pair plots
        of the spectra.

    Parameters
    ----------
    df_row : pd.Series
        An individual row of the DataFrame.

    Returns
    -------
    traces : list of plot.graph_objects
        The bar plot traces for experimental and prosit predicted spectra.
    annotations : list of dict
        A list of the annotations needed for the bar plot.
    """
    index = df_row['index']
    peptide = df_row['peptide']
    pred_intes = [-x for x in df_row['prositIntes']]

    pred_mzs, plotting_names = convert_names_and_mzs(
        df_row['modified_sequence'],
        list(df_row['prositIons'])
    )

    matched_peaks, l2_norm, matched_p_mz, matched_names = experiment_match(
        df_row['mzs'], df_row['intensities'], pred_mzs, plotting_names, mz_accuracy, mz_units
    )
    non_prosit_pred_peaks, prec_peak = get_npp_ions(
        df_row['mzs'], df_row['intensities'], matched_peaks['mzs'], l2_norm, peptide,
        df_row['charge'],
        {
            0: 0.0,
            1: KNOWN_PTM_WEIGHTS['Oxidation (M)'],
            2: KNOWN_PTM_WEIGHTS['Carbamidomethylation'],
        },
        df_row['ptm_seq'], mz_accuracy, mz_units
    )

    unmatched_peaks = get_unmatched(
        df_row['mzs'],
        df_row['intensities'],
        matched_peaks['mzs'],
        non_prosit_pred_peaks['mzs'],
        prec_peak['mzs'],
        l2_norm,
    )

    annotations, colours, extra_traces = get_plot_details(
        index, pred_mzs, matched_p_mz, pred_intes,
        plotting_names, peptide, matched_names, non_prosit_pred_peaks
    )


    for npp_mz, npp_inte, npp_name in zip(
        non_prosit_pred_peaks['mzs'],
        non_prosit_pred_peaks['intensities'],
        non_prosit_pred_peaks['names'],
    ):
        if npp_inte > 0.4:
            if npp_name == 'precursor':
                font_colour = 'plum'
            else:
                font_colour = 'darkseagreen'
            annotations.append(
                {
                    'x': npp_mz,
                    'ax': npp_mz,
                    'y': npp_inte + 0.1,
                    'ay': npp_inte + 0.1,
                    'xref': f'x{index+1}',
                    'yref': f'y{index+1}',
                    'text': npp_name,
                    'font_size': 8,
                    'font_color': font_colour,
                    'showarrow': False,
                }
            )


    traces = create_traces(
        pred_mzs, pred_intes, unmatched_peaks, non_prosit_pred_peaks,
        prec_peak, matched_peaks, colours,
    ) + extra_traces

    return traces, annotations


def plot_spectra(config):
    """ Function to generate pair plots of selected PSMs (experimental vs. Prosit
        predicted spectra.).

    Parameters
    ----------
    config : inspire.config.Config
        The Config object used to run the inSPIRE experiment.
    """
    add_legend(config.output_folder, config.experiment_title, 'standard')

    input_df = pd.read_csv(f'{config.output_folder}/plotData.csv')

    get_charge_from_scan_file = not CHARGE_KEY in input_df.columns
    scan_df = fetch_scan_data(input_df, config, get_charge_from_scan_file)
    scan_df = scan_df.to_pandas()

    input_df = pd.merge(
        input_df,
        scan_df,
        how='inner',
        on=[SOURCE_KEY, SCAN_KEY]
    )

    input_df['pepLen'] = input_df['peptide'].apply(len)
    input_df = input_df[(input_df['pepLen'] > 6) & (input_df['pepLen'] < 31)]
    input_df = input_df[input_df['charge'] < 7]

    n_groups = input_df.shape[0] // PLOTS_PER_PAGE
    if input_df.shape[0] % PLOTS_PER_PAGE:
        n_groups += 1

    input_df['modified_sequence'] = input_df['modifiedSequence'].apply(
        lambda x : x.replace('[+16.0]', '(ox)').replace('[+57.0]', '')
    )
    input_df['precursor_charge'] = input_df[CHARGE_KEY]

    if 'collisionEnergy' in input_df.columns:
        input_df = input_df.rename(columns={'collisionEnergy': 'collision_energy'})
    else:
        input_df['collision_energy'] = config.collision_energy

    input_df[['modified_sequence', 'precursor_charge', 'collision_energy']].to_csv(
        f'{config.output_folder}/plotInput.csv', index=False,
    )

    predict_spectra(config, pipeline='plotSpectra')
    prosit_df = msp_to_df(
        f'{config.output_folder}/plotPredictions.msp', 'prosit', None
    )
    prosit_df = prosit_df.to_pandas()
    prosit_df = prosit_df.drop_duplicates(subset=['modified_sequence', CHARGE_KEY])

    input_df = pd.merge(
        input_df,
        prosit_df,
        how='inner',
        on=['modified_sequence', CHARGE_KEY]
    )

    input_df = input_df.reset_index(drop=True)
    input_df['group'] = input_df.index // PLOTS_PER_PAGE
    input_df['index'] = input_df.index % PLOTS_PER_PAGE
    input_df['ptm_seq'] = input_df['modifiedSequence'].apply(
        convert_mod_seq_to_ptm_seq
    )
    input_df['plot_data'] = input_df.apply(
        lambda x : pair_plot(x, config.mz_accuracy, config.mz_units),
        axis=1,
    )

    if SPECTRAL_ANGLE_KEY not in input_df.columns:
        input_df['results'] = input_df.apply(
            lambda x : calculate_spectral_features(
                x,
                {
                    0: 0.0,
                    1: KNOWN_PTM_WEIGHTS['Oxidation (M)'],
                    2: KNOWN_PTM_WEIGHTS['Carbamidomethylation'],
                },
                config.mz_accuracy,
                config.mz_units,
                None,
                '1',
                config.delta_method,
                minimal_features=True,
            ),
            axis=1,
        )
        input_df['spectralAngle'] = input_df['results'].apply(lambda x : x.get('spectralAngle'))

    titles = []
    for idx in range(input_df.shape[0]):
        seq = input_df[PEPTIDE_KEY].iloc[idx]
        scan_nr = input_df[SCAN_KEY].iloc[idx]
        charge = input_df[CHARGE_KEY].iloc[idx]

        if config.plot_spectra_source_split is not None:
            src = input_df[SOURCE_KEY].iloc[idx].split(config.plot_spectra_source_split)[-1]
        else:
            src = input_df[SOURCE_KEY].iloc[idx]
        spectral_angle = round(input_df[SPECTRAL_ANGLE_KEY].iloc[idx], 2)
        if 'Context' in input_df.columns:
            context = input_df['Context'].iloc[idx]
            titles.append(
                f'<b>Source</b> {src} <b>Scan</b> {scan_nr}<br><b>Peptide</b> {seq} ' +
                f'<b>Charge</b> {charge} <b>Spectral Angle</b> {spectral_angle}<br>' +
                f'<b>Context</b> {context}'
            )
        else:
            titles.append(
                f'<b>Source</b> {src} <b>Scan</b> {scan_nr}<br><b>Peptide</b> {seq} ' +
                f'<b>Charge</b> {charge} <b>Spectral Angle</b> {spectral_angle}<br>'
            )

    for group_idx in range(n_groups):

        start_idx = PLOTS_PER_PAGE*group_idx
        sub_df = input_df[input_df['group'] == group_idx]
        n_plots = sub_df.shape[0]
        n_plot_rows = 1 + (n_plots//PLOTS_PER_LINE)

        fig = make_subplots(
            rows=n_plot_rows,
            cols=PLOTS_PER_LINE,
            subplot_titles = titles[start_idx:start_idx+n_plots],
        )
        for idx in range(1, (n_plot_rows*PLOTS_PER_LINE)+1):
            fig.update_layout(
                {
                    f'xaxis{idx}':{'title_text': 'm/z'},
                    f'yaxis{idx}':{'title_text': 'L<sup>2</sup> Normalized Intensity'},
                }
            )

        plot_data = sub_df['plot_data'].tolist()
        for idx, (traces, annotations) in enumerate(plot_data):
            for trace in traces:
                fig.add_trace(
                    trace,
                    row=1 + (idx//PLOTS_PER_LINE),
                    col=1 + (idx%PLOTS_PER_LINE),
                )

            fig.layout['annotations'] += tuple(annotations)

        for idx in range(sub_df.shape[0]):
            fig.add_trace(
                go.Scatter(
                    x=[0, 1500],
                    y=[0, 0],
                    mode='lines',
                    line={'width':0.5, 'color':'black'},
                ),
                row=1 + (idx//PLOTS_PER_LINE),
                col=1 + (idx%PLOTS_PER_LINE),
            )

        fig = update_fig_layout(fig, n_plot_rows)

        if not config.silent_execution:
            fig.show()

        pio.write_image(
            fig,
            f'{config.output_folder}/spectralPlots{group_idx}.pdf',
            engine='kaleido',
        )

    merger = PdfMerger()
    merger.append(
        f'{config.output_folder}/spectralPlots_legend.pdf'
    )
    for group_idx in range(n_groups):
        merger.append(
            f'{config.output_folder}/spectralPlots{group_idx}.pdf'
        )
    merger.write(f'{config.output_folder}/spectralPlots.pdf')
    merger.close()

    for group_idx in range(n_groups):
        os.remove(
            f'{config.output_folder}/spectralPlots{group_idx}.pdf'
        )
    os.remove(
        f'{config.output_folder}/spectralPlots_legend.pdf'
    )



""" Functions for loading experimental spectra from mzml files.
"""
import numpy as np
import polars as pl
from pyteomics import mzml

from inspire.constants import (
    CHARGE_KEY,
    INTENSITIES_KEY,
    MZS_KEY,
    RT_KEY,
    SCAN_KEY,
    SOURCE_KEY,
)


def process_mzml_file(mzml_filename, scan_ids, with_charge=False, with_retention_time=False):
    """ Function to process an MzML file to find matches with scan IDs.

    Parameters
    ----------
    mzml_filename : str
        The mzml file from which we are reading.
    scan_ids : list of int
        A list of the scan IDs we require.

    Returns
    -------
    scans_df : pd.DataFrame
        A DataFrame of scan results.
    """
    ion_list = []
    intensities_list = []
    scan_id_list = []
    mzml_filenames = []
    filename = mzml_filename.split('/')[-1]

    if with_charge:
        charge_list = []

    if with_retention_time:
        rt_list = []

    with mzml.read(mzml_filename) as reader:
        for spectrum in reader:
            scan_id = int(spectrum['id'].split('scan=')[1])
            if scan_ids is None or scan_id in scan_ids:
                mzml_filenames.append(filename[:-5])
                scan_id_list.append(scan_id)
                intensities_list.append(np.array(list(spectrum['intensity array'])))
                ion_list.append(np.array(list(spectrum['m/z array'])))

                if with_charge:
                    charge_list.append(
                        int(spectrum['precursorList']['precursor'][0]['selectedIonList'][
                            'selectedIon'
                        ][0]['charge state'])
                    )

                if with_retention_time:
                    rt_list.append(float(spectrum['scanList']['scan'][0]['scan start time']))

    scans_df =  pl.DataFrame(
        {
            SOURCE_KEY: pl.Series(mzml_filenames),
            SCAN_KEY: pl.Series(scan_id_list),
            INTENSITIES_KEY: pl.Series(intensities_list),
            MZS_KEY: pl.Series(ion_list)
        }
    )
    if with_charge:
        scans_df[CHARGE_KEY] = pl.Series(charge_list)
    if with_retention_time:
        scans_df[RT_KEY] = pl.Series(rt_list)

    scans_df = scans_df.unique(subset=[SOURCE_KEY, SCAN_KEY])

    return scans_df




from setuptools import setup

setup(
    name='inspirems',
    version='2.0',
    description='Helping to integrate Spectral Predictors and Rescoring.',
    author='John Cormican, Juliane Liepe, Martin Pasen',
    author_email='juliane.liepe@mpinat.mpg.de',
    packages=[
        'inspire',
        'inspire.input',
        'inspire.quant',
        'inspire.pepseek',
        'inspire.plot_spectra',
    ],
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    py_modules=[
        'inspire',
        'inspire.input',
        'inspire.quant',
        'inspire.epitope',
        'inspire.plot_spectra',
    ],
    entry_points={
        'console_scripts': [
            'inspire=inspire.run:run_inspire'
        ]
    },
    python_requires='>=3.11',
    install_requires=[
        'absl-py==2.0.0',
		'astunparse==1.6.3',
		'biopython==1.83',
		'cachetools==5.3.2',
		'certifi==2023.11.17',
		'charset-normalizer==3.3.2',
		'contourpy==1.2.0',
		'cycler==0.12.1',
        'docker==7.0.0',
		'flatbuffers==23.5.26',
		'fonttools==4.45.1',
		'gast==0.5.4',
		'google-auth==2.23.4',
		'google-auth-oauthlib==1.1.0',
		'google-pasta==0.2.0',
		'grpcio==1.59.3',
		'h5py==3.10.0',
		'idna==3.4',
		'joblib==1.3.2',
        'kaleido==0.2.1',
		'keras==2.15.0',
		'kiwisolver==1.4.5',
		'libclang==16.0.6',
		'lxml==4.9.3',
        'logomaker==0.8',
		'Markdown==3.5.1',
		'MarkupSafe==2.1.3',
		'matplotlib==3.8.2',
		'ml-dtypes==0.2.0',
		'numpy==1.26.2',
		'oauthlib==3.2.2',
		'opt-einsum==3.3.0',
		'packaging==23.2',
		'pandas==2.1.3',
		'patsy==0.5.3',
		'Pillow==10.1.0',
		'plotly==5.18.0',
		'polars==0.19.15',
		'protobuf==4.23.4',
        'pyarrow==14.0.1',
		'pyasn1==0.5.1',
		'pyasn1-modules==0.3.0',
		'pyparsing==3.1.1',
		'PyPDF2==3.0.1',
		'pyteomics==4.6.3',
		'python-dateutil==2.8.2',
		'PyYAML==6.0.1',
		'requests==2.31.0',
		'requests-oauthlib==1.3.1',
		'rsa==4.9',
		'scikit-learn==1.3.2',
		'scipy==1.11.4',
		'seaborn==0.13.0',
		'six==1.16.0',
		'statsmodels==0.14.0',
		'tenacity==8.2.3',
		'tensorboard==2.15.1',
		'tensorboard-data-server==0.7.2',
		'tensorflow==2.15.0',
		'tensorflow-estimator==2.15.0',
		'termcolor==2.3.0',
		'threadpoolctl==3.2.0',
		'typing_extensions==4.8.0',
		'tzdata==2023.3',
		'urllib3==2.1.0',
		'Werkzeug==3.0.1',
		'wrapt==1.14.1',
		'xgboost==2.0.2',
        'xlsxwriter==3.1.9',
    ],
    project_urls={
        'Homepage': 'https://github.com/QuantSysBio/inSPIRE',
        'Tracker': 'https://github.com/QuantSysBio/inSPIRE/issues',
    },
)



""" Functions for calculating basic features that do not require spectral data,
    retention time predictors, or binding affinity.
"""
import polars as pl

from inspire.constants import (
    MASS_DIFF_KEY,
    PEPTIDE_KEY,
    PTM_ID_KEY,
    PTM_IS_VAR_KEY,
    PTM_SEQ_KEY,
    SEQ_LEN_KEY,
)
from inspire.spectral_features import fetch_mod_weight_dict

def _check_reps(peptide):
    """ Function to count the number of repeated residues
        within a peptide.

    Parameters
    ----------
    peptide : str
        The peptide sequence.

    Returns
    -------
    n_reps : int
        The number of repeating residues.
    """
    n_reps = 0
    for idx in range(len(peptide)-1):
        if peptide[idx] == peptide[idx+1]:
            n_reps+=1
    return n_reps

def _add_modification_weights(df_row, mod_weight_dict):
    """ Function to add the wight of modifications to a peptide's average residue weight.

    Parameters
    ----------
    df_row : dict or pl.Series
        Colection including ptm_seq, avgResidueMass, and sequenceLength.
    mod_weight_dict : dict
        A dictionary mapping modification IDs to their molecular weight.

    Returns
    -------
    averageResidueWeight : float
        A value for the mean weight of residues including modifications.
    """
    if not isinstance(df_row[PTM_SEQ_KEY], str):
        return df_row['avgResidueMass']

    mod_wt_av = sum(
        mod_weight_dict.get(
            aa_char, 0.0
        ) for aa_char in df_row[PTM_SEQ_KEY]
    )/df_row[SEQ_LEN_KEY]

    return df_row['avgResidueMass'] + mod_wt_av

def create_basic_features(search_df, mods_df):
    """ Function to create the basic features used by percolator/mokapot.

    Parameters
    ----------
    search_df : pd.DataFrame
        A DataFrame of ms search results.

    Returns
    -------
    search_df : pd.DataFrame
        The input DataFrame with update columns containing features
        for percolator/mokapot.
    """
    seq_len_mean = pl.mean(search_df[SEQ_LEN_KEY])

    search_df = search_df.with_columns(
        (pl.col(SEQ_LEN_KEY) - seq_len_mean).abs().alias('seqLenMeanDiff')
    )

    search_df = search_df.with_columns(
        pl.col(MASS_DIFF_KEY).abs().alias('absMassDiff')
    )

    if PTM_IS_VAR_KEY in mods_df.columns:
        mods_df = mods_df[mods_df[PTM_IS_VAR_KEY]]

    if not mods_df.empty:
        mod_ids = [str(x) for x in mods_df[PTM_ID_KEY].tolist()]
        search_df = search_df.with_columns(
            pl.col(PTM_SEQ_KEY).apply(
                lambda x : len(
                    [char for char in x if char in mod_ids]
                ) if isinstance(x, str) else 0,
                skip_nulls=False,
            ).alias('nVarMods')
        )

        mod_weight_dict = fetch_mod_weight_dict(mods_df)
        search_df = search_df.with_columns(
            pl.struct([PTM_SEQ_KEY, SEQ_LEN_KEY, 'avgResidueMass']).apply(
                lambda x : _add_modification_weights(x, mod_weight_dict),
                skip_nulls=False,
            ).alias('avgResidueMass')
        )
    else:
        search_df = search_df.with_columns(
            pl.lit(0).alias('nVarMods')
        )

    search_df = search_df.with_columns(
        pl.col(PEPTIDE_KEY).apply(_check_reps).alias('nRepeatedResidues')
    )

    search_df = search_df.with_columns(
        pl.col(PEPTIDE_KEY).apply(lambda x : len(set(x))/len(x)).alias('fracUnique')
    )

    search_df = search_df.with_columns(
        pl.col(PEPTIDE_KEY).apply(lambda x : x.count('C')/len(x)).alias('fracC')
    )

    search_df = search_df.with_columns(
        pl.col(PEPTIDE_KEY).apply(
            lambda x : (x.count('K') + x.count('R'))/(len(x) - 1)
        ).alias('fracKR')
    )


    return search_df
